/*
 * xsk.c --- sk.c + skcache.c
 *      by k-chinen@is.aist-nara.ac.jp, 2000, 2001.
 *
 * $Id$
 */
/*
 * sk.c --- skip list procedures
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */
/*
 * skcache.c --- object cache procedures using skip list and its bucket
 *              for KOTETU
 *      by k-chinen@is.aist-nara.ac.jp, 2000
 */

/*

 terminology:
  entry   - object entry. store properties like PCB of TCP/IP
            or super block of filesystem.
  frag    - fragment of object cache. object cache is devided little block
            and stored its list.
            you can access entity of object by drawing its list.
  bucket  - list of entry.  it was controled as skip list.
 
*/
/*
 NOTE:

 hash:
    this program use meny hash to assign object to structures in several level
    and identify among objects.

 on the memory:
    cache consists set of buckets and frag lists in entry.

        bucket[0]
            entry[0]    -> frag[0], frag[1], frag[2] ...
            entry[1]
            entry[2]
            ...

        bucket[1]
            entry[0]
            entry[1]
            entry[2]
            ...

 on the disk:
    cache consists dirctories according to hash.

        aa/bb/cc-dd
        
        aa - first level dir. with first hash (so it is bucket)
        bb - second level dir. with second hash
        cc-dd - file with third and fourth hash.

 */

/*
 * See also:
 *   William Pugh, "skiplist cookbook",
 *   UMIACS-TR-89-72.1, July 1989
 */

/*
 * The algorithm on the original text was described by Pascal like
 * langauage.  Subscription were 1..maxlevel or 1..level .
 * Since this is C program,  This file describes subscription as
 * 0..maxlevel-1 and 0..level-1,
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <pthread.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/uio.h>

#include <dirent.h>

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include <sys/mman.h>

#include <assert.h>
#include <errno.h>


#include "range.h"
#include "msg.h"
#include "misc.h"
#include "mem.h"

#include "xsk.h"
#include "sim_stat.h"
#include "sim_hist.h"
#include "mlist.h"

#ifndef MAX
#define MAX(a,b) (((a)>(b))?(a):(b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

#ifndef DISK_RETRY
#define DISK_RETRY (13)
#endif

#ifndef FRAG_NTRY
#define FRAG_NTRY   (32)
#endif

/* object file unit size */
#ifndef OFILE_UNIT
#define OFILE_UNIT  (KB(32))
#endif
#ifndef OFILE_SLOT
#define OFILE_SLOT  (509)
#endif
#ifndef OFILE_TRY
#define OFILE_TRY   (OFILE_SLOT/10)
#endif

#ifndef PACKED_FILENAME0
#define PACKED_FILENAME0    '0'
#endif
#ifndef PACKED_FILENAME 
#define PACKED_FILENAME     "0-0"
#endif
#ifndef INDEX_FILENAME0
#define INDEX_FILENAME0     'I'
#endif
#ifndef INDEX_FILENAME  
#define INDEX_FILENAME      "INDEX"
#endif


#ifndef TUP_MAX
/*
#define TUP_MAX (64)
*/
#define TUP_MAX (4096)
#endif

#define H4_DIR_COUNT_MAX    (128)
#define H4_DIR_SIZE_MAX     (128*K(10))
/*
#define H4_DIR_SIZE_MAX     (K(52))
*/

volatile int dir_count_max      = H4_DIR_COUNT_MAX;
volatile off_t dir_size_max     =  H4_DIR_SIZE_MAX;
volatile off_t dir_size_hwater  =  H4_DIR_SIZE_MAX;
volatile off_t dir_size_lwater  =  H4_DIR_SIZE_MAX/2;


/*
 * arround 256, follwing numbers are known as prime numbers.
 *  241 251 257 263 269 271
 *
 */

int radix_h1 =   241;
int radix_h2 =   239;
int radix_h3 =   233;
int radix_h4 =   251;

skcache_entry_t *NIL;
int skcache_bucket_num;
bucket_t *skcache_bucket_base;

extern pthread_mutex_t stat_lock;

volatile int swapload_count     = 0;
volatile int swapload_success   = 0;
volatile int swapin_count       = 0;
volatile int swapin_success     = 0;
volatile int swapout_count      = 0;
volatile int swapout_success    = 0;

volatile int checkdir_dir_count         = 0;
volatile int checkdir_file_read_count   = 0;
volatile int checkdir_file_handle_count = 0;
volatile int checkdir_file_left_count   = 0;
volatile int checkdir_file_delete_count = 0;
volatile int checkdir_file_erase_count  = 0;

volatile int check_dir_ignore_hash_count = 0;

sim_stat_rec    swapload_time;
static long     swapload_time_count[K(1)];
sim_stat_rec    swapin_time;
static long     swapin_time_count[K(1)];
sim_stat_rec    swapout_time;
static long     swapout_time_count[K(1)];
sim_stat_rec    compaction_time;
static long     compaction_time_count[K(2)];
sim_stat_rec    checkdir_time;
static long     checkdir_time_count[K(2)];
sim_stat_rec    checkdir_time_short;
static long     checkdir_time_short_stat_count[K(2)];
sim_stat_rec    checkmem_time;
static long     checkmem_time_count[K(2)];

sim_stat_rec    swapout_fail_reason;
static long     swapout_fail_reason_count[K(1)];
sim_stat_rec    swapload_fail_reason;
static long     swapload_fail_reason_count[K(1)];



int require_radix_h2;
int require_radix_h3;
off_t require_dir_size;
int require_swap_token_val;
int require_swap_token_interval;
int require_disk_sweep_interval;
int require_keeptime_token_val = -1;



extern pthread_mutex_t config_lock;

int skcache_keeptime_token_init = 10;
int skcache_keeptime_token      = 10;
int skcache_mem_keep_min    = 12;
int skcache_mem_keep_max    = 3600*4;   /* 4 hour, 1/6 day */
int skcache_mem_keep_cur    = 3600;
int skcache_entry_keep_min  = 120;
int skcache_entry_keep_cur  = 86400;    /* 1 day */
int skcache_entry_keep_max  = 86400*7;  /* 1 week */

double skcache_frag_hwater_ratio = 80.0;
double skcache_frag_lwater_ratio = 60.0;

int skcache_signature_refresh = DSIG_REFRESH;

volatile long long swap_count = 0LL;
volatile long long swap_size = 0LL;

/* limitter */
volatile int swapout_token_init = 1000;
volatile int swapout_token  = 1000;
volatile int swapout_token_interval;
volatile int disk_sweep_interval;

time_t min_date;

pthread_mutex_t dir_mutex = PTHREAD_MUTEX_INITIALIZER;
int dir_num=0;
int *dir_lock;
u_int32_t *dir_sig;
off_t *dir_sz;
int *dir_count;
int *dir_time;
int dir_conflict = 0;

#define SENTINEL0   (3249583)
#define SENTINEL1   (7397839)

pthread_mutex_t frag_lock = PTHREAD_MUTEX_INITIALIZER;
skcache_frag_t *z_frag_freelist_dummy[K(1)] = {NULL};
skcache_frag_t *z_frag_freelist = NULL;
int frag_dm0[K(1)] = {SENTINEL0, SENTINEL1};
int frag_use = -1, frag_free = -1;
int frag_dm1[K(1)] = {SENTINEL0, SENTINEL1};
int frag_hhwater_dummy[K(1)];
int frag_hhwater, frag_hwater, frag_lwater;
int frag_lwater_dummy[K(1)];

sim_hist_rec fratio_hist;
static long fratio_hist_area[128];


#define NINDEX_DREAD_MAX    (64)
int nindex_dread=19;

#if 1
/* nanosleep is useful, because it is MT-safe and supports very short time.
 * however, nsec is too short for network I/O.
 * then, msleep is made.
 */
/*
inline
*/
static
int
msleep(int x)
{
    struct timespec w;
    long long tw;

    /* convert msec to nsec */
    tw = x * 1000000;
    w.tv_nsec = tw % 1000000000;
    w.tv_sec =  tw / 1000000000;

    nanosleep(&w, NULL);
    
    return 0;
}
#endif



void
sk_rawdump(FILE* fp, skiplist_t *list, char* (fn)(void*))
{
    skiplist_cell_t *cur;
    int c, i;

#if 0
    for(i=0;i<list->level;i++) {
        fprintf(stdout, " h %4d: %p\n", i, list->header->skip[i]);
    }
#endif

    fprintf(fp, "--- list addr %p, num %d, func %p\n",
        list, list->num, fn);

    c = 0;
    cur = list->header->skip[0];
    if(fn)
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p %s // %p\n",
                c, cur, cur->entity, fn(cur->entity), cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
    else
        while(cur != list->term) {
            fprintf(fp, " %4d: %p : %p // %p\n",
                c, cur, cur->entity, cur->skip[0]);
            cur = cur->skip[0];
            c++;
        }
}


void
sk_dump(skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int k;
    char tmp[L_BUF];
    char line[L_BUF];

    if(list==NULL) {
        return;
    }

    fprintf(fp, "skiplist %p, num %d, level %d/%d\n",
        list, list->num, list->level, list->maxlevel);

    line[0] = '\0';
    sprintf(tmp, "%-38s %-10s %-10s ", "header", "element", "entity");
    strcat(line, tmp);
    for(i=0;i<=list->level;i++) {
        if(list->header->skip[i]==NULL) {
            sprintf(tmp, "%-10s ", "NIL");
        }
        else
        if(list->header->skip[i]==list->term) {
            sprintf(tmp, "%-10s ", "TERM");
        }
        else {
            sprintf(tmp, "0x%08x ", list->header->skip[i]);
        }
        strcat(line, tmp);
    }
    /*
    fputs(line, fp);
    */
    fprintf(fp, "%s\n", line);

    cur = list->header->skip[0];
    while(cur && cur!=list->term) {
        line[0] = '\0';
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
        }
        else {
            strcpy(label, "*NONE*");
        }

        sprintf(tmp, "%-38s 0x%08x 0x%08x ", label, cur, cur->entity);
        strcat(line, tmp);

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                sprintf(tmp, "%-10s ", "NIL");
            }
            else
            if(cur->skip[i]==list->term) {
                sprintf(tmp, "%-10s ", "TERM");
            }
            else {
                sprintf(tmp, "0x%08x ", cur->skip[i]);
            }
            strcat(line, tmp);
        }
        /*
        fputs(line, fp);
        */
        fprintf(fp, "%s\n", line);

        if(cur==list->term) {
            break;
        }
        if(cur==cur->skip[0]) {
            fprintf(fp, "*** looped\n");
            break;
        }

        cur = cur->skip[0];
    }

    fprintf(fp, "%-38s 0x%08x\n", "TERM", list->term);

#if 0
    fprintf(fp, "%-24s %-10s ", "NIL", "NIL");
    fprintf(fp, "\n\n");
#endif


}

void
bk_dump(bucket_t *bk, skiplist_t* list, FILE*fp, char* (fn)(void*) )
{
    char *pcur;
    skiplist_cell_t* cur;
    int i;
    char label[256];
    char *p;
    int j, k;

    if(list==NULL) {
        return;
    }

    fprintf(fp, "### list %p, num %d, level %d/%d, term %p\n",
        list, list->num, list->level, list->maxlevel, list->term);

#ifndef DEBUG
    return;
#endif

#ifdef DEBUG
    fprintf(fp, "#  first %p last %pd\n#\n",
        bk->cell_first, bk->cell_last);

#if 0
    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %s %p\n",
            "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s ",
            "", "", "");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %s %p\n",
            "", "", "", "", "header", list->header);
        fprintf(fp, "#%6s %8s %8s %-14s ",
            "", "", "", "");
    }
    if(list->level<0) {
        fprintf(fp,"*** no level ***\n");
    }
    else {

        cur = list->header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
    fprintf(fp, "\n");

    }
#endif

#if 1
    fprintf(fp, "# cup  header %p: ", bk->cup.header);
        cur = bk->cup.header;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
    fprintf(fp, "# cup  term   %p: ", bk->cup.term);
        cur = bk->cup.term;
        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "NIL");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");

#endif


    fprintf(fp, "#\n");

    if(fn==NULL) {
        fprintf(fp, "#%6s %8s %8s %8s",
            "num", "addr", "entity", "skips..");
    }
    else {
        fprintf(fp, "#%6s %8s %8s %-14s %8s",
            "num", "addr", "entity", "label", "skips..");
    }
    fprintf(fp, "\n");

    j = -1;
    k = 0;
    pcur = (char*)bk->cell_first;
    while(pcur<=(char*)bk->cell_last) {
        cur=(skiplist_cell_t*)pcur;

        j++;
        if(cur->entity==NULL) {
            goto next;
        }
#if 0
        fprintf(fp, "%p %p ", cur->entity, cur);
#endif
        if(fn!=NULL) {
            p = fn(cur->entity);
            if(p) {
                strcpy(label, p);
            }
            else {
                strcpy(label, "*strange*");
            }
            fprintf(fp, "= %6d %6d %p %p %-14s ",
                j, k, cur, cur->entity, label);
        }
        else {
            fprintf(fp, "= %6d %6d %p %p ",
                j, k, cur, cur->entity);
        }

        for(i=0;i<=list->level;i++) {
            if(cur->skip[i]==NULL) {
                fprintf(fp, "%-8s ", "x");
            }
            else {
                fprintf(fp, "%p ", cur->skip[i]);
            }
        }
        fprintf(fp, "\n");
        k++;
#if 0
        if(k>list->num+3) {
            break;
        }
#endif

next:
        pcur += bk->elemsize;
    }

    fprintf(fp, "\n\n");
    fflush(fp);

#endif
}




static
int
randomlevel(int max, double p)
{
    int lv;
    int w = 4097;
    double v;

    lv = 0;
#if 0
    while( v=((double)(rand()%w))/w, v < p && lv < max) {
        lv++;
    }
#endif
    while(drand48() < p && lv < max) {
        lv++;
    }

    return lv;
}




/*
 * cmp - compare
 */
skiplist_cell_t *
sk_find_listpos(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    skiplist_cell_t *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i];
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x;
    }

out:
    return ret;
}

void*
sk_find(skiplist_t *list, void* newface,
        int (cmp)(void*,void*))
{
    register skiplist_cell_t *x;
    register int i, diff;
    void *ret;

    ret = NULL;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[0]==NULL || x->skip[i]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        /* bingo */
        if(diff==0) {
            ret = x->skip[i]->entity;
            goto out;
        }
    }
    x = x->skip[0];

    if(x!=list->term && cmp(x->entity, newface)==0) {
        ret = x->entity;
    }

out:
    return ret;
}


/*
 * if alreay assigned, not new allocation.
 * cmp - compare
 */
void*
sk_assign(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*),
        int* st, int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    skiplist_cell_t *ret;
    int c;
    int diff;

#if 1
    if(newface==NULL) {
        Error("don't assign NULL to skiplist\n");

        ret = NULL;
        *st = -1;
        goto out;
    }
#endif

    if(*flist==NULL) {
        fprintf(stderr, "sk_assign: no free cell in entity list (opt %d)\n", opt);
        ret = NULL;
        *st = -1;
        goto out;
    }

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term
                        || x->skip[i]->entity==NULL) {
                diff = INT_MAX;
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);

            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0) {
        ret = x->entity;
        *st = 1;
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            ret = NULL;
            *st = -1;
            goto out;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        ret = x->entity;
        *st = 0;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

out:
    return ret;
}

int
sk_insert(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (ovw)(void**,void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    register skiplist_cell_t *x;
    register int i;
    int newlevel;
    skiplist_cell_t *nextfree;
    int c;
    int diff;

    x = list->header;
    for(i=list->level; i>=0; i--) {
        c = 0;

        while(1) {
            if(x->skip[0]==NULL || x->skip[0]==list->term) {
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
            c++;
        }
        update[i] = x;
    }
    x = x->skip[0];

assert(x && x->entity);

    if(cmp(x->entity, newface)==0 && ovw) {
        ovw(&x->entity, newface);
    }
    else {
        newlevel = randomlevel(list->maxlevel, 0.5);

assert(newlevel<=list->maxlevel);

        if(newlevel > list->level) {
            for(i=list->level+1;i<=newlevel;i++) {
                update[i] = list->header;
            }
            list->level = newlevel;
        }

        if(*flist==NULL) {
            fprintf(stderr, "no free cell (opt %d)\n", opt);
            return -1;
        }

        nextfree = (*flist)->skip[0];

        x = *flist;
        *flist = nextfree;

        x->entity = newface;
        list->num++;

        for(i=0;i<=newlevel;i++) {
            x->skip[i] = update[i]->skip[i];
            update[i]->skip[i] = x;
        }

    }

    return 0;
}

/*
 * arguments:
 *      cmp - compare
 *      rem - remove
 * return:
 *      NULL    fail    
 *      !NULL   success
 */
void*
sk_delete(skiplist_t *list, void* newface,
        skiplist_cell_t **flist,
        int (cmp)(void*,void*), int (rem)(void*), int opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x;
    int i;
    void *ret;
    int diff;
    int chk;
    int z;
    char* skcache_label4_ref(void *pp);

    ret = NULL;

    x = list->header;
    z = 0;
    for(i=list->level; i>=0; i--) {
        while(1) {
            if(x->skip[i]==NULL || x->skip[i]==list->term) {
                break;
            }
            if(x->skip[i]->entity==NULL) {
                Error("sk_delete: found empty entity");
#if 1
    sk_dump(list, stderr, skcache_label4_ref);
#endif
                break;
            }
            diff = cmp(x->skip[i]->entity, newface);
            if(diff>=0) {
                break;
            }
            x = x->skip[i];
        }
        update[i] = x;
    }
    x = x->skip[0];

    if(x==NULL) {
Error("sk_delete: meat NULL element");
        return ret;
    }
    if(x==list->term) {
Error("sk_delete: reach to end");
        return ret;
    }
    if(x->entity==NULL) {
Error("sk_delete: meat NULL entity");
        return ret;
    }

    if(x->entity == (skcache_entry_t*)newface
      || cmp(x->entity, newface)==0) {
        if(rem) {
            chk = rem(x->entity);
            if(chk) {
Error("sk_delete: rem() return !0 (%d)", chk);
                return ret;
            }
        }

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x) {
                update[i]->skip[i] = x->skip[i];
            }
        }

        while(list->level>0 &&
                list->header->skip[list->level]==list->term) {
            list->level--;
        }

        if(x->entity==NULL) {
Error("sk_delete: entity is NULL");
        }

        ret = x->entity;
        x->entity = NULL;
        for(i=1;i<=list->maxlevel;i++) {
            /*
            x->skip[i] = NULL;
            */
            x->skip[i] = list->term;
        }
        x->skip[0] = *flist;
        *flist = x;

        list->num--;
    }
#if 1
    else {
        Trace("sk_delete: not found entry %p\n", newface);
    }
#endif

#if 0
    if(ret==NULL) {
Error("sk_delete: return is NULL");
    }
#endif

    return ret;
}

/*
 * ack - decide delete or not
 *          that function have to return not zero for delete
 */
int
sk_mapdelete(skiplist_t *list,
        skiplist_cell_t **flist,
        int (ack)(void*, void*), void *opt)
{
    skiplist_cell_t *update[SKL_MAXLEVEL];
    skiplist_cell_t *x, *y, *n;
    int i;
    int ret;
    int diff;
    int chk;

    ret = 0;

    x = list->header;
    for(i=0;i<=list->level;i++) {
            update[i] = x;
    }

    x = x->skip[0];

    while(x!=list->term) {
        n = x->skip[0]; /* save next element */

        if(ack==NULL)
            goto delete;

        chk = ack(x->entity, opt);

        if(chk) {
delete:
            /*
             * skip current element
             * so, it is mean deletation of element
             * this routine don't care actual enitity procedure
             * (e.g, free(), unlink() or something )
             */
            for(i=0;i<=list->level;i++) {
                if(update[i]->skip[i]==x) {
                    update[i]->skip[i] = x->skip[i];
                }
            }

            /* add current element to free-list for recycle */
            y = x;
            y->entity = NULL;
            for(i=1;i<=list->maxlevel;i++) {
#if 0
                y->skip[i] = NULL;
#endif
                y->skip[i] = list->term;
            }
            y->skip[0] = *flist;
            *flist = y;

            list->num--;

            ret++;
        }
no_delete:

        for(i=0;i<=list->level;i++) {
            if(update[i]->skip[i]==x)
                update[i] = x;
        }

        x = n;
    }

    return ret;
}

skiplist_t*
sk_new(int nlevel, void* user_term)
{
    skiplist_t *this;
    skiplist_cell_t *n;
    int i, j;

    this = (skiplist_t*)MALLOC(sizeof(skiplist_t)*1);
    if(this==NULL) {
        return NULL;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    this->maxlevel  = nlevel;
    this->level     = 0;

    n = this->header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)MALLOC(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    n = this->term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)MALLOC(
                    sizeof(void*)+sizeof(skiplist_cell_t**)
                    +sizeof(skiplist_cell_t*)*(nlevel+1));
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = this->term;
        }

    return this;
}


int
bucket_show(bucket_t *bk)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;

    fprintf(stdout, "=== bk %p\n", bk);

    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    n  = bk->cup.header->skip[0];
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    fprintf(stdout, "free list: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    i  = 0;
    while(n) {
        fprintf(stdout, "%4d: <%p> %p %p\n",
            i, n, n->entity, n->skip[0]);
        n = n->skip[0];
        i++;
    }
    fprintf(stdout,"  have %d item(s)\n", i);

    return 0;
}

int
bk_new(bucket_t *bk, int nlevel, int nelement, void* user_term)
{
    skiplist_cell_t *n;
    skiplist_cell_t **p;
    int i, j;
    char *pool;


    if(bk==NULL) {
        return 0;
    }

    if(nlevel>SKL_MAXLEVEL-1) {
        nlevel = SKL_MAXLEVEL-1;
    }

    bk->elemnum = nelement;
    bk->elemsize = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(nlevel+1);
    bk->totalsize = bk->elemsize*(nelement+4);

    pool = (char*)MALLOC(bk->totalsize);
    if(pool==NULL) {
        fprintf(stderr, "not enough memory !!!\n");
        return -1;
    }

#if 0
    fprintf(stderr, "nlevel %d, nelement %d\n",
            nlevel, nelement);
    fprintf(stderr, "element size %d, num %d -> total %d\n",
            bk->elemsize, nelement, bk->totalsize);
    fprintf(stderr, "pool  [%p..%p)\n", pool, pool+bk->totalsize);
#endif
#if 0
    for(i=0;i<nelement+2;i++) {
        fprintf(stdout, "%4d: <%p>\n", i, pool+bk->elemsize*i);
    }
    fflush(stdout);
#endif
    if(pthread_mutex_init(&bk->mutex, NULL)<0) {
        Error("fail pthread_mutex_init() (%d)", errno);
    }

    bk->cup.num         = 0;
    bk->cup.maxlevel    = nlevel;
    bk->cup.level       = 0;
    bk->cup.header      = (skiplist_cell_t*)pool;
    bk->cup.term        = (skiplist_cell_t*)(pool + 1*bk->elemsize);
    n = bk->cup.header;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }
    n = bk->cup.term;
        n->entity = user_term;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=0;j<=nlevel;j++) {
            n->skip[j] = bk->cup.term;
        }


#if 0
    fprintf(stderr, "cup  header %p\n", bk->cup.header);
    fprintf(stderr, "cup  term   %p\n", bk->cup.term);
#endif

    bk->elem_freelist = (skiplist_cell_t*)(pool+4*bk->elemsize);
#if 0
    fprintf(stderr, "elem_freelist    %p\n", bk->elem_freelist);
#endif


#if 0
#ifndef NDEBUG
    bk->cell_first = (skiplist_cell_t*)(pool+0*bk->elemsize);
    bk->cell_last  = (skiplist_cell_t*)(pool+nelement*bk->elemsize);

    fprintf(stderr, "  first..last %p..%p\n", bk->cell_first, bk->cell_last);
#endif
#endif

    n = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL) {
            break;
        }
#if 0
        fprintf(stdout, "* %d, %p\n", i, n);
        fflush(stdout);
#endif
        n->entity = NULL;
        n->skip = (skiplist_cell_t**)((char*)n+sizeof(void*)*2);
        for(j=1;j<=nlevel;j++) {
/*
            n->skip[j] = NULL;
*/
            n->skip[j] = bk->cup.term;
        }
        if(i<nelement-1) {
            n->skip[0] = (skiplist_cell_t*)((char*)n+bk->elemsize);
            n = (skiplist_cell_t*)((char*)n+bk->elemsize);
        }
        else {
            n->skip[0]  = NULL;
        }
    }

#if 0
    fprintf(stdout, "cup  header,skip: %p\n", bk->cup.header->skip);
    fprintf(stdout, "free: %p\n", bk->elem_freelist);

    n  = bk->elem_freelist;
    for(i=0;i<nelement;i++) {
        if(n==NULL)
            break;

#if 0
        fprintf(stdout, "%4d: <%p> ", i, n);
        for(j=0;j<=nlevel;j++) {
            fprintf(stdout, "%d:<%p> %p ",
                j, &(n->skip[j]), n->skip[j]);
        }
        fprintf(stdout, "\n");
#endif
#if 1
        fprintf(stdout, "%4d: <%p> %p\n", i, n, n->skip[0]);
#endif

        n = n->skip[0];
    }

#endif

    return bk->totalsize;
}


/*
 * return
 *      1   fine
 *      0   strange
 */
int
sk_emptyentity_check(bucket_t *bk)
{
    int ret;
    skiplist_t *list;
    register skiplist_cell_t *x;

    ret = 1;
    list = &bk->cup;
    x = list->header;
    x = x->skip[0];
    while(x!=list->term) {
        if(x->entity == NULL) {
            Error("sk_emptyentity_check: bk# %d, elem %p have empty entityd",
                bk->bid, x);
            ret = 0;
            goto out;
        }
        x = x->skip[0];
    }

out:
    return ret;
}


/*
 * return
 *      1   fine
 *      0   strange
 */
int
sk_entity_check(bucket_t *bk, char *f, int l)
{
    int ret;
    skiplist_t *list;
    register skiplist_cell_t *x;
    register int c, n, tn;

    ret = 1;
    c = 0;
    n = 0;
    tn = 0;
    list = &bk->cup;
    x = list->header;
    x = x->skip[0];
    while(x!=list->term) {
        if(x->entity == NULL) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have empty entity :%s %d",
                bk->bid, c, x, f, l);
            ret = 0;
            n++;
        }
        else
        if(((skcache_entry_t*)x->entity)->issue_time < 0) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have ignore issue_time %d :%s %d",
                bk->bid, c, x, ((skcache_entry_t*)x->entity)->issue_time, f, l);
            ret = 0;
            n++;
        }
        else
        if(((skcache_entry_t*)x->entity)->oid < 0) {
            Error(
"sk_entity_check: bk# %d, pos %d, elem %p have ignore oid# %d :%s %d",
                bk->bid, c, x, ((skcache_entry_t*)x->entity)->oid, f, l);
            ret = 0;
            n++;
        }

        if(((skcache_entry_t*)x->entity)->oid == 0) {
            tn++;
        }

        x = x->skip[0];
        c++;
    }
    if(n>0) {
        Trace("sk_entity_check: bk# %d, found %d ignore entities: %s %d\n",
            bk->bid, n, f, l);
    }
    if(tn>0) {
        Trace("sk_entity_check: bk# %d, found %d temporary entities: %s %d\n",
            bk->bid, tn, f, l);
    }

out:
    return ret;
}






#include "xsk_hash.c"
extern int calc_hash(unsigned char *b, int *h1, int *h2, int *h3, int *h4);


void
dsig_mkhumanstr(u_int32_t *s, char *line)
{
    int i;
    u_int32_t m, v;
    char *p;

    p = line;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        m = 1U << i%(SIZE_OF_UINT*8);
        v = s[i/(SIZE_OF_UINT*8)] & m;
        *p++ = (v ? '1' : '0');
    }
    *p = '\0';
}

void
dsig_dump(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */

    dsig_mkhumanstr(s, line);
    Trace("%s\n", line);
}

void
dsig_show(u_int32_t *s)
{
    char line[DSIG_BIT_WIDTH+2];    /* keep more 2 chars for \n + \0 */
    int i, c;

    dsig_mkhumanstr(s, line);
    c = 0;
    for(i=0;i<DSIG_BIT_WIDTH;i++) {
        if(DSIG_TMP_ISSET(s, i)) {
            c++;
        }
    }
    Trace("addr %p count %d, bits %s\n", s, c, line);
}


void
dsig_set(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    s[i/(SIZE_OF_UINT*8)] |= m;
}

int
dsig_isset(u_int32_t *s, int i)
{
    u_int32_t m;
    m = 1UL << i%(SIZE_OF_UINT*8);
    return (s[i/(SIZE_OF_UINT*8)] & m) ? 1 : 0;
}

int
dsig_allset(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)-1L; /* all bits were set */
        s[i] = u;
    }
}

int
dsig_allclear(u_int32_t *s)
{
    int i;
    u_int32_t u;

    for(i=0;i<DSIG_ELEM_WIDTH;i++) {
        u = (u_int32_t)0UL;
        s[i] = u;
    }
}




int
skcache_entry_show(skcache_entry_t* e, FILE *ofp)
{
    fprintf(ofp,"\
-/-/\n\
entry %p, oid# %d, last# %d, hash %d:%d:%d:%d\n\
name %s\n\
count %d, ref %d, cracked %d, owner %p\n\
issue %d, hot %d, lmd %d, expire %d\n\
totallength %lld\n\
mem record %p, disk record %d\n\
\n",
        e, e->oid, e->last_oid,
        e->h1, e->h2, e->h3, e->h4,
        e->name,
        e->count, e->ref, e->cracked,
        e->owner, 
        e->issue_time, e->hot, e->lmd, e->expire,
        (long long)e->totallength,
        e->m_recorded, e->d_recorded);
    
    return 0;
}


int
skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            if(x->h3==y->h3) {
                return strcmp((const char*)x->name, (const char*)y->name);
            }
            else {
                return x->h3 - y->h3;
            }
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

int
X_skcache_entry_cmp(void *px, void *py)
{
    skcache_entry_t *x = (skcache_entry_t*)px;
    skcache_entry_t *y = (skcache_entry_t*)py;

    if(x->h1==y->h1) {
        if(x->h2==y->h2) {
            return x->h3 - y->h3;
        }
        else {
            return x->h2 - y->h2;
        }
    }
    else {
        return x->h1 - y->h1;
    }
}

char*
skcache_label_hash(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_PATH];

    sprintf(tmp, "%04d:%04d:%04d", p->h1, p->h2, p->h3);

    return tmp;
}

char*
skcache_label_short(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d %s", p->h1, p->h2, p->h3, p->name);

    return tmp;
}

char*
skcache_label_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d-%03d", p->h1, p->h2, p->h3, p->ref);

    return tmp;
}

char*
skcache_label4_ref(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%04d:%04d:%04d:%04d-%03d",
        p->h1, p->h2, p->h3, p->h4, p->ref);

    return tmp;
}

char*
skcache_label_long(void *pp)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    static char tmp[L_URL*2];

    sprintf(tmp, "%03d:%03d:%03d:%04d=%06d-%03d-0x%08x",
        p->h1, p->h2, p->h3, p->h4, p->oid, p->ref, p->owner);

    return tmp;
}

int
skcache_count_frag_in_entry(skcache_entry_t *en)
{
    int c;
    skcache_frag_t *p;

    c = 0;
    p = en->m_recorded;
    while(p) {
        c++;
        p = p->next;
    }

    return c;
}


int
skcache_dump_entry_frag(char *msg, skcache_entry_t *en)
{
    skcache_frag_t *p;
    off_t frag_sum;
    int u;

    Trace("%s\n", msg);
    Trace("  frags of entry %p, oid# %d,\n", en, en->oid);
    if(en==NULL)
        return -1;

    p = en->m_recorded;
    frag_sum = 0L;
    u = 0;
    while(p) {
        frag_sum += p->len;
        Trace("  %4d: p %p next 0x%08x | len %4d/%d %7ld\n",
            u, p, p->next, p->len, p->size, (long)frag_sum);
        if(p==p->next) {
            Trace("  looped !!!\n");
            break;
        }
        p = p->next;
        u++;
    }
    Trace("  num %d, totalength %lld\n",
        u, (long long)en->totallength);

    return u;
}


off_t
skcache_show_with_HTML(int ofd)
{
    int i;
    off_t wsum, w;
    skiplist_t      *list;
    skiplist_cell_t *cur;
    skcache_entry_t *en;

    wsum = 0;

    w = fdprintf(ofd, "<html>\n");
    wsum += w;



    for(i=0;i<skcache_bucket_num;i++) {
        w = fdprintf(ofd, "<p>bucket# %d\n", i);
        wsum += w;

        w = fdprintf(ofd, "<listing>\n");
        wsum += w;

        list = &skcache_bucket_base[i].cup;
        cur = list->header->skip[0];

        while(cur && cur!= list->term) {
            en = (skcache_entry_t*)cur->entity;
            w = fdprintf(ofd, " %04d:%04d:%04d %d %d '%s'\n",
                    en->h1, en->h2, en->h3, en->ref, en->count, en->name);
            if(w>0) {
                wsum += w;
            }

            cur = cur->skip[0];
        }

        w = fdprintf(ofd, "</listing>\n");
        wsum += w;

    }


    w = fdprintf(ofd, "<html>\n");
    wsum += w;

    Trace("write %ld bytes\n", wsum);
}


int
skcache_have(int fh, int sh, int th, int ff, char *name,
    int flag_check_rcode, time_t ref_time)
{
    int ret;
    skcache_entry_t tmp;
    skcache_entry_t *stored_pos;
    bucket_t *bk;


    ret = 0;

    bk = &skcache_bucket_base[fh];

    tmp.h1 = fh;
    tmp.h2 = sh;
    tmp.h3 = th;
    tmp.h4 = ff;
    strcpy(tmp.name, name);

    LOCK_BK(bk);

    stored_pos = (skcache_entry_t*) sk_find(&bk->cup, &tmp, skcache_entry_cmp);

    if(stored_pos) {
#if 1
        if(strcmp((const char*)stored_pos->name, (const char*)name)!=0) {
            Error("skcache_have: hash collision %04d:%04d:%04d '%s' v.s. '%s'",
                fh, sh, th, stored_pos->name, name);
            goto out;
        }
#endif

        if(flag_check_rcode) {
        }

        if(ref_time>0) {
        }

        ret = 1;
    }
    else {
        ret = 0;
    }

out:
    UNLOCK_BK(bk);

    return ret;
}


static
long long
alloc_entry_freelist(skcache_entry_t** head, int num)
{
    char *base;
    int esize;
    int i;
    int ret;
    skcache_entry_t *pos;
    skcache_entry_t *last;

    esize = sizeof(skcache_entry_t);
    base = (char*)MALLOC(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_entry_t*)(base + esize*i);

        pos->magic[0] = MAGIC_0;
        pos->magic[1] = MAGIC_1;

        pos->oid = -1;
        pos->ref = -1;
        pos->h1 = -1;
        pos->h2 = -1;
        pos->h3 = -1;
        pos->name[0] = '\0';
        pos->m_recorded = NULL;
        pos->d_recorded = -1;

        pos->next = (skcache_entry_t*)((char*)pos + esize);
    }

    *head = (skcache_entry_t*)base;
    pos->next = NULL;

#if 0

    Trace("entry// head %p, num %d\n", head, num);
    Trace("esize %d, base %p\n", esize, base);

    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, name '%s', m_recorded %p ->next %p\n",
                i, pos, pos->name, pos->m_recorded, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}

int
skcache_frag_freelist_check(skcache_frag_t* base)
{
    int i;
    skcache_frag_t *pos;
    int sumsz, sumlen;
    
    pos = base;
    i = 0;
    sumsz = sumlen = 0;
    while(pos) {
        sumsz  += pos->size;
        sumlen += pos->len;
        pos = pos->next;
        i++;
    }
    Trace("skcache_frag_freelist_check: base %p, count %d, length sum %d, av. %.2f %d // size sum %d av. %.2f %d\n",
        base, i,
        sumlen, ((double)sumlen)/i , sumlen%i, 
        sumsz, ((double)sumsz)/i , sumsz%i );

    return i;
}

static
long long
alloc_frag_freelist(skcache_frag_t** head, int num, int sz)
{
    char *base;
    int esize;
    int i;
    skcache_frag_t *pos;
    skcache_frag_t *last;
    int sum;

    esize = sizeof(skcache_frag_t)+sz;
    base = (char*)MALLOC(esize * num);
    if(base==NULL) {
        return 0;
    }

    for(i=0; i<num; i++) {
        pos = (skcache_frag_t*)(base + esize*i);
        pos->size = sz;
        pos->len = 0;
        pos->ent = (char*)((char*)pos+sizeof(skcache_frag_t));

        pos->next = (skcache_frag_t*)((char*)pos+esize);
    }
    *head = (skcache_frag_t*)base;
    pos->next = NULL;

    skcache_frag_freelist_check(*head);

#if 0
    Trace("frag// head %p, num %d, sz %d\n", head, num, sz);
    Trace("esize %d, base %p\n", esize, base);

    i = 0;
    sum = 0;
    pos = (skcache_frag_t*)base;
    while(pos) {
        sum += pos->size;
        pos = pos->next;
        i++;
    }
    Trace("count %d, totalsize %d, av.size %.f\n",
        i, sum, ((double)sum)/i );

#endif

#if 1
    Trace("frag// head %p, num %d, sz %d // total %lld\n",
        head, num, sz, (long long)esize*num);
    Trace("esize %d, base %p\n", esize, base);
#endif

#if 0
    if(num<100) {
        i = 0;
        pos = *head;
        while(pos) {
            Trace(" %2d: pos %p, len %d, sz %d ->next %p\n",
                i, pos, pos->len, pos->size, pos->next);

            pos = pos->next;
            i++;
        }
    }

#endif

    return esize*num;
}

int
echk_count=0;

int
skcache_entry_check(skcache_entry_t* e, char*f, int l)
{
    int ret;
    int cur_count;

    ret = 0;

    if(e==NULL) {
        Error("echk: %s %d : empty entry", f, l);
        ret = -1;
        goto out;
    }
#ifdef CHECK_MAGIC
    if(e->magic[0]!=MAGIC_0 || e->magic[1]!=MAGIC_1) {
        Error("echk: %s %d : entry %p, oid# %d, ignore magic",
            f l, e, e->oid);
        ret = 2;
        goto out;
    }
#endif
#if 0
    if(e->ref<0) {
        Error("echk: %s %d : entry %p, oid# %d, owner %p, neg ref %d",
            f, l, e, e->owner, e->oid, e->ref);
        ret = 3;
        goto out;
    }
#endif
#if 1
    if(e->ref<0) {
        e->ref = 0;
    }
#endif

out:
#if 0
    if(ret) {
        pthread_mutex_lock(&stat_lock);
        echk_count++;
        cur_count = echk_count;
        pthread_mutex_unlock(&stat_lock);
        if(cur_count>3) {
            skcache_entry_show(e, stdout);
            exit(0);
        }
    }
#endif

    return ret;
}

int skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);
int RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf);
static int multi_frag_lock();


int
skcache_entry_purge(bucket_t *bk, skcache_entry_t *e)
{
    skcache_frag_t *cur, *pre;
    int chk;

    chk = multi_frag_lock();
    if(chk) {
        return -1;
    }

    if(e->m_recorded) {
        /* remove */
        cur = e->m_recorded;
        while(cur) {
            pre = cur->next;

            RAW_skcache_push_idle_frag(bk, cur);
            bk->frag_use--;
            if(pre==cur) {
                Error("skcache_entry_purge: bk %p, entry %p, frag %p is looped",
                    bk, e, cur);
                break;
            }

            cur = pre;
        }

        e->m_recorded = NULL;
    }

    MUTEX_UNLOCK(&frag_lock);

    return 0;
}



int
skcache_entry_delete(bucket_t *bk, skcache_entry_t *e)
{
    void *chkv;
    int chk;
    int ret;

    if(ret = skcache_entry_check(e, __FILE__, __LINE__)) {
        Trace("skcache_entry_delete: bk# %d, entry %p is ignore, but delete\n",
            bk->bid, e);
#if 1
        goto out;
#endif
    }
    ret = 0;

#if 0
    Trace("skcache_entry_delete: bk# %d %p, entry %p, oid# %d, ref %d\n",
        bk->bid, bk, e, e->oid, e->ref);
#endif

 assert(sk_emptyentity_check(bk));
    chk = skcache_entry_purge(bk, e);
 assert(sk_emptyentity_check(bk));
    if(chk) {
        Error("skcache_entry_delete: fail skcache_entry_purge() bk# %d, entry %x",
            bk->bid, e);
        ret = 1;
    }
    else {

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
 assert(sk_emptyentity_check(bk));
        chkv = sk_delete(&bk->cup, e,
            &bk->elem_freelist, skcache_entry_cmp, NULL, 0);
 assert(sk_emptyentity_check(bk));
        if(chkv==NULL) {
            Trace("skcache_entry_delete: bk# %d, entry %p is not found\n",
                bk->bid, e);

            ret = 2;

            chkv = sk_find(&bk->cup, e, skcache_entry_cmp);
            Trace("  sk_find() returns %x\n", chkv);

        }
        else {
            chk = skcache_push_idle_entry(bk, e);
#if 1
            if(chk) {
                Trace("skcache_entry_delete: chk %d\n", chk);
                ret = 3;
            }
#endif
        }

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
    }

out:
    return ret;
}


#ifndef HAVE_DIRENT_T
typedef struct dirent dirent_t;
#endif


char* cache_dir_prefix = "/var/tmp/cache";
int radix_disk = -1;


inline
char*
assign_dir_prefix(char *name, int d1)
{
    if(radix_disk<=0) {
        sprintf(name, "%s/%d", cache_dir_prefix, d1);
    }
    else {
        sprintf(name, "%s%d/%d",
            cache_dir_prefix, d1%radix_disk, d1);
    }

    return name;
}


inline
char*
assign_dirname_bynumber(char *name, int d1, int d2, int d3)
{
    char tmp[L_PATH];

    assign_dir_prefix(name, d1);

    sprintf(tmp, "/%d", d2);
    strcat(name, tmp);

    return name;
}

inline
char*
assign_dirname(char* name, skcache_entry_t *e, int level)
{
    assign_dirname_bynumber(name, e->h1, e->h2, e->h3);

    return name;
}

inline
char*
assign_filename(char *name, skcache_entry_t *e)
{
    char tmp[L_PATH];

    assign_dirname(name, e, -1);
    
    sprintf(tmp, "/%d-%d", e->h3, e->h4);
    strcat(name, tmp);

    return name;
}

inline
char*
assign_indexfilename(char *name, skcache_entry_t *e)
{
    char tmp[L_PATH];

    assign_dirname(tmp, e, -1);
    pathncat(name, L_PATH, tmp, INDEX_FILENAME);

    return name;
}



int
build_dir_bynumber(int d1, int d2, int d3)
{
    int i;
    int ret=0;
    int chk;
    char name[L_PATH];
    char tmp[L_PATH];

l1:
    assign_dir_prefix(name, d1);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 1 (%d)", name, errno);
        ret = -1;
        goto out;
    }

l2:
    sprintf(tmp, "/%d", d2);
    strcat(name, tmp);
    chk = mkdir(name, 0755);
    if(chk && errno!=EEXIST) {
        Error("build_dir_bynumber fail mkdir() '%s' 2 (%d)\n", name, errno);
        ret = -1;
        goto out;
    }

out:
    return ret;
}

int
build_dir(skcache_entry_t *e, int lev)
{
    int ret=0;

    ret = build_dir_bynumber(e->h1, e->h2, e->h3);

out:
    return ret;
}



pthread_mutex_t fd_assign_lock = PTHREAD_MUTEX_INITIALIZER;
int *ifd_pool;
int *pfd_pool;
int *pos_pool;


/*
 * return:
 *      0   success
 *      !0  fail and indicate reason
 *
 */
int
ifd_open(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;

    ret = 0;
    *rfd = -1;

#if 0
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_UNIT);
#endif

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = 0; 
    xe->h4 = 0; 

    fpos = FDPOOL_POS_E(xe);

#if 0
    Trace("ifd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)OFILE_UNIT,
        (long)fpos);
#endif

MUTEX_LOCK(&fd_assign_lock);

    sfd = ifd_pool[fpos];
    if(sfd>=0) {
        *rfd = sfd;
#if 0
Trace("ifd_open: cached %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif
        goto out;
    }


    namechk = assign_indexfilename(name, xe);

    c = 0;
try_open:
    c++;
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2, xe->h3);
            goto try_open;
        }
        if(errno==ENOSPC) {
            Error("ifd_open: disk full");
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("ifd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = 37;
        goto fail_open;
    }

    ifd_pool[fpos] = sfd;
    pos_pool[fpos] = 0;
    *rfd = sfd;

#if 0
Trace("ifd_open: new    %4d %d,%d *rfd %d, return %d\n",
    fpos, xe->h1, xe->h2, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;


out:

MUTEX_UNLOCK(&fd_assign_lock);

#if 0
Trace("ifd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

    return ret;
}

int
pfd_open(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;
    int fpos;

    ret = 0;
    *rfd = -1;

#if 0
    Trace("pfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_UNIT);
#endif

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;
    xe->h3 = 0; 
    xe->h4 = 0; 

    fpos = FDPOOL_POS_E(xe);

#if 0
    Trace("pfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld, pos %ld\n",
        e->h1, e->h2, e->h3, e->h4,
        (long)e->totallength, (long)OFILE_UNIT,
        (long)fpos);
#endif

MUTEX_LOCK(&fd_assign_lock);

    sfd = pfd_pool[fpos];
    if(sfd>=0) {
        *rfd = sfd;
#if 0
Trace("pfd_open: cached %d,%d *rfd %d, return %d\n",
    xe->h1, xe->h2, *rfd, ret);
#endif
        goto out;
    }


    namechk = assign_filename(name, xe);

    c = 0;
try_open:
    c++;
/*
    sfd = open(name, O_WRONLY | O_CREAT, 0666);
*/
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2, xe->h3);
            goto try_open;
        }
        if(errno==ENOSPC) {
            Error("pfd_open: disk full on open()");
            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("pfd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = 49;
        goto fail_open;
    }

    pfd_pool[fpos] = sfd;
    *rfd = sfd;

#if 0
Trace("pfd_open: new    %d,%d sfd %d, *rfd %d, return %d\n",
    xe->h1, xe->h2, sfd, *rfd, ret);
#endif

    goto out;

fail_open:
    (void)1;

out:

MUTEX_UNLOCK(&fd_assign_lock);

#if 0
Trace("pfd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

    return ret;
}

/* this routine never called */
int
pfd_close(int *rfd)
{
#if 0
    if(*rfd>=0) {
        close(*rfd);
        *rfd = -1;
    }
#endif
}


int
nfd_open(int *rfd, bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int sfd;
    char name[L_PATH];
    char *namechk;
    int c;
    off_t pos, dst;
    skcache_entry_t _xe;
    skcache_entry_t *xe;

    ret = 0;
    *rfd = -1;

#if 1
    Trace("nfd_open: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_UNIT);
#endif

    /* only use h1 and h2 */
    xe = &_xe;
    *xe = *e;

    namechk = assign_filename(name, xe);

    c = 0;
try_open:
    c++;
/*
    sfd = open(name, O_WRONLY | O_CREAT, 0666);
*/
    sfd = open(name, O_RDWR | O_CREAT, 0666);
    if(sfd<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_open;
        }
        if(errno==ENOENT && c<DISK_RETRY) {
            build_dir_bynumber(xe->h1, xe->h2, xe->h3);
            goto try_open;
        }
        if(errno==ENOSPC) {
            MUTEX_LOCK(&stat_lock);
            Error("nfd_open: disk full on open()");
            swapout_token = 0;
            MUTEX_UNLOCK(&stat_lock);

            ret = XSK_DISKFULL;
            goto fail_open;
        }
        Error("pfd_open: cannot open swap file '%s' %d (%d)", name, c, errno);

        ret = 39;
        goto fail_open;
    }

    *rfd = sfd;

#if 1
Trace("nfd_open: new    %d,%d *rfd %d, return %d\n",
    xe->h1, xe->h2, *rfd, ret);
#endif

seek_phase:
    goto out;

fail_open:
    (void)1;

out:
#if 1
Trace("nfd_open: oid# %d, *rfd %d, return %d\n", xe->oid, *rfd, ret);
#endif

    return ret;
}

int
nfd_close(int *rfd)
{
    if(*rfd>=0) {
        close(*rfd);
        *rfd = -1;
    }
}



#define OFFSET_OFILE(e) ((e->h3+e->h4)%OFILE_SLOT)


int
dump_ifile(int ifd, int start_pos)
{
    int i;
    int n;
    skcache_entry_t tmp_entry;
    off_t pos;
    int nr;
    int diff;

    n = start_pos;
    for(i=0;i<OFILE_TRY;i++) {
        pos = n*sizeof(skcache_entry_t);
        lseek(ifd, pos, SEEK_SET);

        nr = read(ifd, &tmp_entry, sizeof(skcache_entry_t));

        diff = tmp_entry.issue_time - min_date;
    Trace("  :: fd# %d, try %3d: pos %4d %12lldB : magic(hex) %02x %02x, oid# %d, [%d] diff %d\n",
            ifd, i, n, (long long)pos,
            (unsigned char)tmp_entry.magic[0],
            (unsigned char)tmp_entry.magic[1],
            tmp_entry.oid, tmp_entry.h4, diff
        );

        n = (n+1)%OFILE_SLOT;
    }

    return 0;
}


int
pfd_seekfree(int sfd, bucket_t *bk, skcache_entry_t *e, int *rpos)
{
    int ret;
    int c;
    off_t pos, dst;
    int fpos;
    int npos;
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[NINDEX_DREAD_MAX];
    int i, wid;


    ret = 0;
    *rpos = (off_t) -1;
    fpos = FDPOOL_POS_E(e);

#if 0
    Trace("pfd_seekfree: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_UNIT);
#endif

    npos = OFFSET_OFILE(e);
    pos = (off_t)npos * sizeof(skcache_entry_t);

#if 0
Trace("pfd_seekfree: sfd# %d, [%03d:%03d:%03d:%04d] x %6ld => pos %8d/%d %12lldB (%4ldMB+%7ld)\n",
    sfd,
    e->h1, e->h2, e->h3, e->h4, OFILE_UNIT,
    npos, OFILE_SLOT,
    (long long)pos, (long)dM(pos), (long)mM(pos) );
#endif

    c = 0;
try_seek:
    c++;
    dst = lseek(sfd, pos, SEEK_SET);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
Error("pfd_seekfree: cannot seek %lld on index file fd# %d :%d (%d)",
            (long long)pos, sfd, c, errno);

        ret = -76;
        goto out;
    }
    vpos = dst;


    try = OFILE_TRY;
    match = 0;
    do {
        pos = (off_t)npos * sizeof(skcache_entry_t);
        wid = sizeof(skcache_entry_t)
            * (MIN(npos+nindex_dread, OFILE_SLOT) -  npos);

        c = 0;
try_read:
        c++;
        nr = read(sfd, chunk, sizeof(skcache_entry_t));
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("pfd_seekfree: cannot read fd# %d: %d (%d)",
                sfd, c, errno);
            break;
        }
        vpos += nr;

        /* block read */
        
        wid = nr/sizeof(skcache_entry_t);
        for(i=0;i<wid && try;i++) {
            tmp_entry = &chunk[i];

#ifdef TRACE_INDEX
        diff = tmp_entry->issue_time - min_date;
Trace("  :f fd# %d, try %3d: pos %4d %12lldB : vpos %12lldB : magic(hex) %02x %02x, oid# %d, diff %d\n",
        sfd, try, npos, (long long)pos, (long long) vpos,
        (unsigned char)tmp_entry->magic[0], (unsigned char)tmp_entry->magic[1],
        tmp_entry->oid, diff
        );
#endif

            if(tmp_entry->magic[0]!=MAGIC_0 || tmp_entry->magic[1]!=MAGIC_1
                || tmp_entry->oid == 0 || tmp_entry->issue_time < min_date) {
#if 0
                Trace("pfd_seekfree: unmatch MAGIC, it is free\n");
#endif
                match = 1;
                break;
            }
    
            npos = (npos+1)% OFILE_SLOT;
            try--;
        }
        if(match) {
            break;
        }
        
        if(npos==0 && !match) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif
            c = 0;
try_rewind:
            c++;
            dst = lseek(sfd, 0, SEEK_SET);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("pfd_seekfree: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
            }
            vpos = dst;
        }

    } while (match==0 && try>0);

    if(match) {
        c = 0;
try_back:
        c++;
        dst = lseek(sfd, - (vpos - pos), SEEK_CUR);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("pfd_seekfree: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = -39;
            *rpos = npos;
            goto out;
        }
        vpos = dst;

        ret = npos;
        *rpos = npos;
    }
    else {
#if 0
Trace("pfd_seekfree: sfd# %d, not found, [%d:%d:%d:%d] fpos %d, npos %d\n",
        sfd, e->h1, e->h2, e->h3, e->h4, FDPOOL_POS_E(e), OFFSET_OFILE(e));
#endif
        ret = -1;
        *rpos = npos;
    }

    goto out;

out:

#if 0
#endif
#ifdef TRACE_INDEX
    pos = (off_t)*rpos * sizeof(skcache_entry_t);
Trace("pfd_seekfree: [%03d:%03d:%03d:%04d] sfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, sfd, OFILE_TRY - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}


int
pfd_seekobj(int *rfd, bucket_t *bk, skcache_entry_t *e,
    int *rpos, skcache_entry_t *re)
{
    int ret;
    int sfd;
    int c;
    off_t pos, dst;
    int fpos;
    int npos;
    int nr;
    skcache_entry_t *tmp_entry;
    int match = 0;
    int try;
    off_t vpos;
    time_t diff;
    skcache_entry_t chunk[NINDEX_DREAD_MAX];
    int wid;
    int i;

    *rpos = (off_t)-1;
    ret = 0;
    sfd = *rfd;
    fpos = FDPOOL_POS_E(e);
    npos = OFFSET_OFILE(e);
    pos = (off_t)npos * sizeof(skcache_entry_t);

#if 0
    Trace("pfd_seekobj: [%03d:%03d:%03d:%04d] len %ld/%ld\n",
        e->h1, e->h2, e->h3, e->h4, (long)e->totallength, (long)OFILE_UNIT);
#endif

#if 0
    if(pos_pool[fpos]==pos) {
        vpos = pos_pool[fpos];
        goto skip_seek_top;
    }
#endif

    c = 0;
try_seek:
    c++;
    dst = lseek(sfd, pos, SEEK_SET);
    if(dst<(off_t)0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_seek;
        }
        Error("pfd_seekobj: cannot seek %lld on swap file fd# %d :%d (%d)",
            (long long)pos, sfd, c, errno);

        ret = 79;
        goto out;
    }
    vpos = dst;

skip_seek_top:


    try = OFILE_TRY;
    match = 0;
    do {
        pos = (off_t)npos * sizeof(skcache_entry_t);
        wid = sizeof(skcache_entry_t)
            * (MIN(npos+nindex_dread, OFILE_SLOT) -  npos);

        /* block read */

        c = 0;
try_read:
        c++;
#if 0
Trace("seekobj: read sfd %d, pos %d %lld, addr %p, wid %d\n",
    sfd, npos, (long long)pos, chunk, wid);
#endif
        nr = read(sfd, chunk, wid);
        if(nr<0) {
            if(errno==EINTR && c < DISK_RETRY) {
                goto try_read;
            }
            Error("pfd_seekobj: cannot read fd# %d pos %lld: %d (%d)",
                sfd, (long long)pos, c, errno);
        }
        vpos += nr;
        wid = nr/sizeof(skcache_entry_t);

        for(i=0;i<wid && try;i++) {
            pos = (off_t)npos * sizeof(skcache_entry_t);
            tmp_entry = &chunk[i];

#ifdef TRACE_INDEX
            diff = tmp_entry->issue_time - min_date;
Trace("  :o fd# %d, try %3d: pos %4d %12lldB : vpos %12lldB : magic(hex) %02x %02x, oid# %d, h4 %d vs %d, diff %d\n",
            sfd, try, npos, (long long)pos, (long long)vpos,
            (unsigned char)tmp_entry->magic[0],
            (unsigned char)tmp_entry->magic[1],
            tmp_entry->oid,
            tmp_entry->h4, e->h4,
            diff
            );
#endif
            if(tmp_entry->magic[0]==MAGIC_0 && tmp_entry->magic[1]==MAGIC_1) {
                if(tmp_entry->h4<0 || tmp_entry->oid==0) {
                    /* it is empty */
                }
                else
                if(tmp_entry->h4==e->h4 && strcmp(tmp_entry->name,e->name)==0) {
                    *re = *tmp_entry;   /* copy */
                    match = 1;
                    break;
                }
            }

            npos = (npos+1)% OFILE_SLOT;
            try--;
        }
        if(match || try<=0) {
            break;
        }
    
        if(npos==0 && !match) {
#ifdef TRACE_INDEX
            Trace("  : rewind sfd# %d,\n", sfd);
#endif

            c = 0;
try_rewind:
            c++;
            dst = lseek(sfd, 0, SEEK_SET);
            if(dst<(off_t)0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_rewind;
                }
    Error("pfd_seekobj: cannot rewind %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);
                break;
            }
            vpos = dst;
        }
    } while (match==0 && try>0);

    if(match) {
#if 0
Trace("pfd_seekobj: sfd# %d, found, [%d:%d:%d:%d] npos %d, pos %lld, vpos %ld\n",
    sfd, e->h1, e->h2, e->h3, e->h4,
    npos, (long long)pos, (long long) vpos);
#endif

        c = 0;
try_back:
        c++;
        dst = lseek(sfd, - (vpos-pos), SEEK_CUR);
        if(dst<(off_t)0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_back;
            }
    Error("pfd_seekobj: cannot back %lld on index file fd# %d :%d (%d)",
        (long long)pos, sfd, c, errno);

            ret = -1;
            *rpos = npos;
            goto out;
        }
        vpos = dst;

        ret = 0;
        *rpos = npos;

#if 0
Trace(
"pfd_seekobj: sfd# %d, found, [%d:%d:%d:%d] npos %d, pos %lld, vpos %ld\n",
    sfd, e->h1, e->h2, e->h3, e->h4,
    npos, (long long)pos, (long long) vpos);
#endif
    }
    else {
#if 0
Trace("pfd_seekobj: sfd# %d, not found, [%d:%d:%d:%d] fpos %d, npos %d\n",
        sfd, e->h1, e->h2, e->h3, e->h4, FDPOOL_POS_E(e), OFFSET_OFILE(e));
#endif
        ret = -1;
        *rpos = npos;
    }

    goto out;


out:

#ifdef TRACE_INDEX
    pos = (off_t)*rpos * sizeof(skcache_entry_t);
Trace("pfd_seekobj:  [%03d:%03d:%03d:%04d] *rfd# %d, try %d, match %d, pos %d %12lld <%12lld>, return %d\n",
    e->h1, e->h2, e->h3, e->h4, *rfd, OFILE_TRY - try, match,
    *rpos, (long long) pos, (long long) vpos, ret);
#endif

    return ret;
}




/* 
    return:
        1   fail
        0   success
 */
int
write_attach_fd(int *rifd, int *rpfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int chk;
    int fpos;
    int npos;
    off_t bipos, bppos; 
    int ifd, pfd;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);

    *rifd = *rpfd = -7;
    *rnpos = -1;

    ifd = pfd = -3;

    chk = ifd_open(&ifd, bk, e);
    if(chk) {
        Trace("write_attach_fd: fail open ifd# %d, chk %d\n", ifd, chk);
        *rifd = -1;
        ret = chk;

        goto out;
    }

    *rifd = ifd;


    chk = pfd_seekfree(ifd, bk, e, &npos);
    if(chk<0) {
#if 0
        Trace("write_attach_fd: not found or error (%d)\n", chk);
#endif
        *rnpos = -1;
        ret = 1;

        goto out;
    }
    *rnpos = npos;
    bipos = npos*sizeof(skcache_entry_t);
    bppos = npos*OFILE_UNIT;


    chk = pfd_open(&pfd, bk, e);
    if(chk) {
        Error("write_attach_fd: fail open pfd# %d, chk %d", pfd, chk);
        *rpfd = -1;
        ret = 1;

        goto out;
    }

    *rpfd = pfd;

    c = 0;
try_lseek:
    c++;
    chk = lseek(pfd, bppos, SEEK_SET);
    if(chk<0) {
Error("write_attach_fd: fail lseek at npos %ld %lldB in pfd# %d, chk %d (%d)",
(long)npos, (long long)bppos, pfd, chk, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_lseek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
        ret = 0;
    }


out:
#if 0
Trace("write_attach_fd: done *rifd# %d, *rpfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rpfd, (long)*rnpos, (long long)bppos, ret);
#endif

    return ret;
}

/* 
    return:
        1   fail
        0   success
 */

int
read_attach_fd(int *rifd, int *rpfd, int *rnpos,
    bucket_t *bk, skcache_entry_t *e, skcache_entry_t *re)
{
    int ret;
    int chk;
    off_t vpos;
    int fpos;
    int npos;
    int ifd, pfd;
    off_t bpos;
    int c;

    ret = 0;
    fpos = FDPOOL_POS_E(e);


    *rifd = *rpfd = -25;
    *rnpos = -1;
    
    ifd = pfd = -57;

    chk = ifd_open(&ifd, bk, e);
    if(chk) {
        Trace("read_attach_fd: fail open ifd# %d, chk %d\n", ifd, chk);
        *rifd = -1;
        ret = chk;

        goto out;
    }

    *rifd = ifd;

    chk = pfd_seekobj(&ifd, bk, e, &npos, re);
    if(chk) {
#if 0
        Trace("read_attach_fd: not found\n");
#endif
        *rnpos = -1;
        ret = 1;

        goto out;
    }

    *rnpos = npos;
    bpos = npos*OFILE_UNIT;

    chk = pfd_open(&pfd, bk, e);
    if(chk) {
Trace("read_attach_fd: fail open pfd# %d, chk %d\n", pfd, chk);

        *rpfd = -1;
        ret = chk;

        goto out;
    }
        
#if 0
    {
        off_t v, v2;
        v = lseek(ifd, 0, SEEK_CUR);
Trace("read_attach_fd: v %d %lldB on ifd# %d %d\n",
    npos, (long long)v, ifd, *rifd);

        dump_ifile(ifd, OFFSET_OFILE(e));

        v2 = lseek(ifd, v, SEEK_SET);
    }
#endif


    *rpfd = pfd;

    c = 0;
try_lseek:
    c++;
    vpos = lseek(pfd, bpos, SEEK_SET);
    if(vpos<0) {
Error("read_attach_fd: fail lseek at %ld %lld in pfd# %d, vpos %d (%d)",
    (long)npos, (long long)bpos, pfd, vpos, errno);
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_lseek;
        }
        *rnpos = -1;
        ret = 1;
    }
    else {
#if 0
Trace("read_attach_fd: done lseek at pos %ld %lldB in pfd# %d, vpos %lldB\n",
    (long)npos, (long long)bpos, pfd, (long long)vpos);
#endif
        ret = 0;
    }


out:
#if 0
Trace("read_attach_fd: done *rifd# %d, *rpfd# %d, pos %ld %lldB, ret %d\n",
        *rifd, *rpfd, (long)*rnpos, (long long)bpos, ret);
#endif

    return ret;
}




int
detach_fd(int *rfd)
{
#ifndef OBJECT_PACKING
    /*
    pfd_close(rfd);
    */
#else
    (void)1;
#endif
}


/*
 * return
 *      0   success
 *      >0  fail
 *
 */

int
skcache_entry_swapload(bucket_t *bk, skcache_entry_t *e)
{
    int ifd, pfd;
    int npos;
    skcache_entry_t tmp_entry;
    skcache_frag_t *cur, *pre;
    int nr;
    off_t nr_sum;
    int chk;
    int i;
    int ret;
    int nf;
    struct timeval begin, end;
    int c;
    off_t fsize;
    off_t tmpfsize;
    int erase;
    int sigpos;
    int lf;


    gettimeofday(&begin, NULL);
    ret = 0;

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif

#if 0
    Trace("swapload: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength);
#endif


    if(e->m_recorded!=NULL) {
        Error("swapload: not recorded in memory bk# %d, e %p", bk->bid, e);
        ret = 2;
        goto out;
    }

    sigpos = DSIG_POS(e->h1, e->h2, e->h3);

    MUTEX_LOCK(&stat_lock);
    swapload_count++;
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&stat_lock);

    if(lf>0) {
/*
        Trace("swapload: conflict directory bk# %d, dir# %d,.\n", bk, sigpos);
*/
        MUTEX_LOCK(&dir_mutex);
        dir_conflict++;
        MUTEX_UNLOCK(&dir_mutex);

        ret = 73;
        goto out;
    }

#if 0
    chk = read_attach_fd(&ifd, &pfd, &npos, bk, e);
#endif

    chk = read_attach_fd(&ifd, &pfd, &npos, bk, e, &tmp_entry);
    if(chk || ifd<0 || pfd<0 || ifd==pfd || npos<0) {
#if 0
Trace("swapload: not found object (%d) or error ifd# %d, pfd# %d, pos %ld\n",
    chk, ifd, pfd, (long)npos);
#endif
        ret = 4;
        goto out;
    }
#if 0
    else {
        Trace("swapload: found object in ifd# %d,\n", ifd);
    }
#endif

#if 0
    Trace("swapload: found object in ifd# %d, pfd# %d, \n", ifd, pfd);
    {
        off_t v;
        v = lseek(ifd, 0, SEEK_CUR);
        Trace("swapload: ifd# %d, v %lldB\n",
            ifd, (long long)v);
    }
#endif


#if 0
    c = 0;
read_cntl:
    c++;
    nr = read(ifd, &tmp_entry, sizeof(skcache_entry_t));
    if(nr<0) {
        if(errno==EINTR && c < DISK_RETRY) {
            goto read_cntl;
        }
        Error("swapload: cntl, fail read() return %d %c (%d)", nr, c, errno);
        ret = 123;
        goto fail_read_cntl;
    }

    if(nr<sizeof(skcache_entry_t)) {
#if 0
        Trace("swapload: read control short than control %ld/%ld\n",
            (long)nr, sizeof(skcache_entry_t));
#endif
        ret = 126;
        goto fail_read_cntl;
    }

#endif



    /*
     * this routing MUST check magic.
     * because disk maybe stores broken or older version cache file. 
     */
    if(tmp_entry.magic[0]!=MAGIC_0
        || tmp_entry.magic[1]!=MAGIC_1) {
#ifdef CHECK_MAGIC
        Error("swapload: ignore magic %02x %02x / %02x %02x '%s'",
            tmp_entry.magic[0], tmp_entry.magic[1],
            MAGIC_0, MAGIC_1, tmp_entry.name);
#endif
        ret = 125;
        goto fail_read_cntl;
    }

    tmp_entry.m_recorded = NULL;
#if 0
    Trace("swapload: read ifd# %d, npos %d, oid# %d, [%d:%d:%d:%d] %s\n",
        ifd, npos, 
        tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
        tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
    Trace("  target is [%d] '%s'\n", e->h4, e->name);
#endif

    if(tmp_entry.oid<=0) {
#if 0
        Trace("swapload: reject negative oid in %p, oid# %d,\n",
            &tmp_entry, tmp_entry.oid);
#endif
        ret = 127;
        goto fail_read_cntl;
    }

    if(tmp_entry.h4<0) {
#if 0
        Trace("swapload: reject negative h4 in %p, h4 %d,\n",
            &tmp_entry, tmp_entry.h4);
#endif
        ret = 126;
        goto fail_read_cntl;
    }

    if(strcmp(tmp_entry.name,e->name)!=0) {
#if 0
        Trace("swapload: read oid# %d, hash %d,%d,%d,%d %s\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4, tmp_entry.name);
        Trace("but target is '%s'\n", e->name);
#endif

#if 0
        Trace("    oid# %d, hash %d,%d,%d,%d\n",
            tmp_entry.oid, tmp_entry.h1, tmp_entry.h2,
            tmp_entry.h3, tmp_entry.h4);
        Trace("    saved %s\n    wont  %s\n", tmp_entry.name, e->name);
        {
            int yh1, yh2, yh3, yh4;
            calc_hash(tmp_entry.name,   &yh1, &yh2, &yh3, &yh4);
            calc_hash(e->name,          &yh1, &yh2, &yh3, &yh4);
        }
#endif

        ret = 133;
        erase = 1;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<0) {
#if 1
        Error("swapload: ignore issue_time");
#endif
        ret = 134;
        goto fail_read_cntl;
    }

    if(tmp_entry.issue_time<min_date) {
#if 1
        Error("swapload: very old object");
#endif
        ret = 129;
        goto fail_read_cntl;
    }

    {
        int old_oid;
        void *old_owner;

        old_oid = e->oid;
        old_owner = (void*)e->owner;

    *e = tmp_entry;

        (void*)e->owner = old_owner;
        e->d_recorded = e->oid;
    }
    
skip_read_cntl:
    (void)1;





    pre = NULL;

    nr_sum = (off_t)0L;
    nf = 0;     /* the number of frag */
    do {
        cur = skcache_pull_new_frag(bk);
        if(cur==NULL) {
            ret = 7;
            goto fail_read;
        }
        bk->frag_use++;

        c=0;
try_read:
        c++;
        nr = read(pfd, cur->ent, cur->size);
#if 0
        Trace("<<< %3d %p: read %5d/%ld %7ld? oid#%d, %lld\n",
            nf, cur, nr, (long)cur->size,
            (nr>0) ? (long)nr_sum+nr : (long)nr_sum,
            e->oid, (long long)e->totallength);
#endif
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read;
            }

            Error("swapload: data, fail read() return %d %d (%d)",
                nr, c, errno);

            cur->len = 0;
            skcache_push_idle_frag(bk, cur);
            bk->frag_use--;
            cur = NULL;

            ret = 6;
            goto fail_read;
        }
        if(nr==0) {
#if 1
            Trace("  oid# %d, reach end\n", e->oid);
#endif
            cur->len = 0;

            /* This frag is not first frag, remove that to save frags */
            if(nf>0) {
                skcache_push_idle_frag(bk, cur);
                bk->frag_use--;
                cur = NULL;
            }
            
            break;
        }

        if(nr_sum + nr > e->totallength) {
#if 0
            Trace("swapload: cut oid# %d, sum %lld + nr %lld > total %lld\n",
                e->oid,
                (long long)nr_sum, (long long)nr, (long long)e->totallength);
#endif

            cur->len = e->totallength - nr_sum;
            nr_sum += e->totallength - nr_sum;
        }
        else {
#if 0
            Trace("swapload: add oid# %d, sum %lld + nr %lld <= total %lld\n",
                e->oid,
                (long long)nr_sum, (long long)nr, (long long)e->totallength);
#endif

            cur->len = nr;
            nr_sum += (off_t)nr;
        }

        if(pre==NULL) {
            if(e->m_recorded) {
Error("oid# %d, over write %p upon %p", e->oid, cur, e->m_recorded);
            }
            e->m_recorded = cur;
        }
        else {
            pre->next = cur;
        }

        nf++;
        pre = cur;
        cur->next = NULL;   /* dummy */
    } while( nr_sum < e->totallength );
 
    detach_fd(&pfd);

#if 0
    Trace("  swapload read %lld/%lld %lld oid# %d, '%s'\n",
        (long long)nr_sum, (long long)e->totallength,
        (long long)tmp_entry.totallength,
        e->oid, e->name);
#endif

    MUTEX_LOCK(&stat_lock);
    swapload_success++;
    MUTEX_UNLOCK(&stat_lock);

    goto out;


/*
 * recoverely
 */

fail_read_cntl:
#if 0
    Trace("swapload: fail read contorl block reason %d, erase %d, size %ld\n",
        ret, erase, (long)fsize);
#endif
    goto fail;

fail_read:
#if 0
    Trace("swapload: fail to read data. trush just readings, reason %d\n",
        ret);
#endif
    goto fail;

fail:
    detach_fd(&pfd);
   
    sim_stat_add_value(&swapload_fail_reason, ret);

#if 0
    if(erase) {
        unlink(name);
        MUTEX_LOCK(&stat_lock);
        swap_count--;
        swap_size -= fsize;
        MUTEX_UNLOCK(&stat_lock);
    }
#endif

    goto out;



out:
#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif

    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapload_time, KOTETU_SUB_TIME(end, begin));

#if 0
    if(ret==0) {
        chk = sk_entity_check(bk, __FILE__, __LINE__);
    }
#endif
#if 0
    Trace("swapload: bk# %d, entry %p, oid# %d, return %d\n",
        bk->bid, e, e->oid, ret);
#endif
#if 0
    Trace("swapload: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %d, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, e->totallength, ret);
#endif

    return ret;
}




int
skcache_entry_swapin(bucket_t *bk, skcache_entry_t *e)
{
    return skcache_entry_swapload(bk, e);
}










/*
 *
 * return
 *      0   success
 *      !0  fail
 *              >0  ignore
 *              <0  temporary, redo again
 */

int
skcache_entry_swapout(bucket_t *bk, skcache_entry_t *e)
{
    int ret;
    int ifd, pfd;
    int npos;
    skcache_frag_t *cur, *pre;
    int nw;
    off_t nw_sum;
    int chk;
    int i;
    struct timeval begin, end;
    struct stat sbuf;
    int sig_pos;
    int c;

    struct iovec iov[HARD_NFRAG_DWRITE];
    int     numf;
    off_t   lenf;

    off_t last_size;
    int lf;
    int sigpos;
    int left_token;


    gettimeofday(&begin, NULL);

    ret = 0;

    if(bk==NULL) {
        Error("nil bucket pointer %x", bk);
        ret = 79;
        goto fail_start;
    }
    if(e==NULL) {
        Error("nil entry pointer %x", e);
        ret = 77;
        goto fail_start;
    }


#if 0
    Trace("swapout: b bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength);
#endif

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif

    if(ret = skcache_entry_check(e, __FILE__, __LINE__)) {
        ret = 147;
        goto out;
    }

    if(e->m_recorded==NULL) {
      Trace("swapout: not recorded in memory bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
        ret = 0;
        goto out;
    }
    if(e->issue_time==0 || e->oid==0) {
#if 0
        Trace("swapout: temporary entry bk# %d, e %p, oid# %d, length %ld\n",
            bk->bid, e, e->oid, (long)e->totallength);
#endif
        ret = 0;
        goto out;
    }

#ifdef NO_SAVECOUNT
    if(e->d_recorded!=0)
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        ret = 0;
        goto out;
    }
#endif

#if 0
    if(e->totallength+sizeof(skcache_entry_t) > OFILE_UNIT)
#endif
#if 1
    if(e->totallength > OFILE_UNIT)
    {
#if 0
        Trace("swapout: oid# %d, file size %ld is larger than %ld\n",
            e->oid,
            (long)e->totallength+sizeof(skcache_entry_t), (long)OFILE_UNIT);
#endif
        ret = 0;
        goto out;
    }
#endif


    sigpos = DSIG_POS(e->h1, e->h2, e->h3);

    MUTEX_LOCK(&stat_lock);
    left_token = swapout_token;
    if(left_token>0) {
        swapout_token--;
        swapout_count++;
    }
    lf = dir_lock[sigpos];
    MUTEX_UNLOCK(&stat_lock);

    if(left_token<=0) {
        ret = -999;
        goto fail_start;
    }
    else
    if(left_token==1) {
        Trace("swapout: *last* token of bk# %d, %p\n",
            bk->bid, bk);
    }

    if(lf>0) {
/*
        Trace("swapout: conflict directory bk# %d, dir# %d,.\n",
            bk->bid, sigpos);
*/
        MUTEX_LOCK(&dir_mutex);
        dir_conflict++;
        MUTEX_UNLOCK(&dir_mutex);

        ret = -777;
        goto fail_start;
    }


    chk = write_attach_fd(&ifd, &pfd, &npos, bk, e);
    if(chk || ifd<0 || pfd<0 || ifd==pfd || npos<0) {
/*
*/
#if 0
        Trace(
         "swapout: not found object (%d) or error ifd# %d, pfd# %d, pos %ld\n",
            chk, ifd, pfd, (long)npos);

        Trace("  dir# %d, dir_sz %lld %.2f%% hwater/max=%lld/%lld\n",
            sigpos, (long long)dir_sz[sigpos],
            ((double)dir_sz[sigpos]*100.0/(double)dir_size_max),
            (long long)dir_size_hwater, (long long)dir_size_max
            );
#endif

        ret = 84;
        goto fail_open;
    }

#if 0
    Trace("swapout: oid# %d, attach to file ifd# %d, pfd# %d, \n",
        e->oid, ifd, pfd);
#endif

    /***
     *** write control
     ***/

    c = 0;
try_write_cntl1:
    c++;
    nw = write(ifd, e, sizeof(skcache_entry_t));
    if(nw<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_write_cntl1;
        }

        if(errno==ENOSPC) {
            Error("disk full!! at write control in swapout 1");
        }
        else
        if(errno==ENOMEM) {
            Error("kernel said no memory. reset token <%d>", c);

            MUTEX_LOCK(&stat_lock);
            swapout_token = 0;
            MUTEX_UNLOCK(&stat_lock);

            msleep(50);

            if(c<DISK_RETRY) {
                goto try_write_cntl1;
            }

        }
        else {
            Error("fail write() 1 return %d %d (%d)", nw, c, errno);
        }
        ret = 5;
        goto fail_write_cntl;
    }

#if 0

    c = 0;
try_write_cntl2:
    c++;
    nw = write(pfd, e, sizeof(skcache_entry_t));
    if(nw<0) {
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_write_cntl2;
        }

        if(errno==ENOSPC) {
            Error("disk full!! at write control in swapout 2");
        }
        else
        if(errno==ENOMEM) {
            Error("kernel said no memory. reset token <%d>", c);

            MUTEX_LOCK(&stat_lock);
            swapout_token = 0;
            MUTEX_UNLOCK(&stat_lock);

            msleep(50);

            if(c<DISK_RETRY) {
                goto try_write_cntl2;
            }

        }
        else {
            Error("fail write() 2 return %d %d (%d)", nw, c, errno);
        }
        ret = 5;
        goto fail_write_cntl;
    }

#endif


    if(e->d_recorded!=0) 
    {
#if 0
    Trace("+# oid# %d, d_recorded# %d, count %3d,\n\thot %d, name %s\n",
        e->oid, e->d_recorded, e->count, e->hot, e->name);
#endif
        goto skip_body_swapout;
    }


    /***
     *** write body
     ***/

    cur = e->m_recorded;
    if(cur==NULL) {
        goto skip_body_swapout;
    }

    nw_sum = 0L;
    i = 0;
    while(cur) {

        /* write frag chunk by writev() */
        numf = 0;
        lenf = (off_t)0L;
        while(cur && numf<nfrag_dwrite) {
#if 0
            Trace(">>> %3d %3d %p: stored %5d/%d %7ld\n",
                i, numf, cur->ent, cur->len, cur->size, (long)nw_sum+cur->len);
#endif
            if(cur->len>0) {
                iov[numf].iov_base  = cur->ent;
                iov[numf].iov_len   = cur->len;
                numf++;
                lenf += (off_t) cur->len;
            }
            cur = cur->next;
        }

        c = 0;
try_write:
        c++;
        nw = writev(pfd, iov, numf);
#if 0
        Trace(">>> %3d write %ld\n",
            i, nw);
#endif

        if(nw<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_write;
            }
            Error("fail write() return %d (%d)", nw, errno);
            if(errno==ENOSPC) {
                Error("disk full!! when swapout data");
            }

            ret = 7;
            goto fail_write;
        }
        else
        if(nw==0) {
            break;
        }
        if(nw>0) {
            nw_sum += nw;
        }

        i++;
    }
#if 0
    Trace("swapout: oid# %d, swap out %lld bytes/ %lld bytes\n",
        e->oid, (long long)nw_sum, (long long)e->totallength);
#endif

    /* XXX */
#if 0
    ftruncate(pfd, (off_t) nw_sum + sizeof(skcache_entry_t) );
#endif

skip_body_swapout:

    detach_fd(&pfd);
    skcache_entry_purge(bk, e);
    
    e->m_recorded = NULL;
    e->d_recorded = e->oid;
#if 1
    e->swapout_candidate = 0;
#endif

#if 0
    Trace("swapout: oid# %d, write %ld, %ld+%ld/%ld\n",
        e->oid, sizeof(skcache_entry_t) + (long)nw_sum,
        sizeof(skcache_entry_t), (long)nw_sum, (long)e->totallength);
#endif

#if 0
    Trace("swapout: p %p, oid# %d, path '%s', size %ld\n", 
        e, e->oid, name, (long)nw_sum);
#endif

    sig_pos = DSIG_POS(e->h1, e->h2, e->h3);
    last_size = 0;

    MUTEX_LOCK(&stat_lock);
    swapout_success++;

    if(last_size<=0) {
        swap_count += 1LL;
        swap_size +=
            (long long)(nw_sum + (off_t)sizeof(skcache_entry_t));
    }
    else {
        swap_size +=
            (long long)(nw_sum + (off_t)sizeof(skcache_entry_t) - last_size);
    }
#if 0
    Trace("swapout: oid# %d, total swap count %lld size %lld\n",
        e->oid, swap_count, swap_size);
#endif
    MUTEX_UNLOCK(&stat_lock);


    MUTEX_LOCK(&dir_mutex);

    if(dir_count[sig_pos]<0) {
        /* nothing */
    }
    else {
        DSIG_SET(sig_pos, e->h4/DSIG_VALUE_RES);
        dir_sz[sig_pos] +=
            (off_t)(nw_sum + (off_t)sizeof(skcache_entry_t) - last_size);
        dir_count[sig_pos]++;
    }

    MUTEX_UNLOCK(&dir_mutex);

#if 0
    dsig_show(&dir_sig[sig_pos*DSIG_ELEM_WIDTH]);
#endif

    goto out;

fail_write_cntl:
fail_write:
    detach_fd(&pfd);

    sim_stat_add_value(&swapout_fail_reason, ret);

    MUTEX_LOCK(&stat_lock);
    if(last_size<0) {
    }
    else {
        swap_size -= last_size;
    }
    MUTEX_UNLOCK(&stat_lock);

    goto out;

fail_open:
fail_start:

    sim_stat_add_value(&swapout_fail_reason, ret);

    goto out;

out:
#if 0
    Trace("  frag use/all = %d/%d\n", bk->frag_use, bk->frag_all);
#endif
    gettimeofday(&end, NULL);
    sim_stat_add_value(&swapout_time, KOTETU_SUB_TIME(end, begin));

#if 0
    chk = sk_entity_check(bk, __FILE__, __LINE__);
#endif
#if 0
    Trace("swapout: a bk# %d, entry %p, oid# %d, d_recorded %d, m_recorded %p, count %d, len %ld, return %d\n",
        bk->bid, e, e->oid, e->d_recorded, e->m_recorded, e->count, (long)e->totallength, ret);
#endif

    return ret;
}





int
skcache_entry_purge_check_noref(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;

#if 0
    Trace("  entry_check_noref: p %p ref %d in bk%p ", p, p->ref, bk);
#endif

    if(p->ref>0) {
#if 0
        Trace("refered %d session, not purge\n", p->ref);
#endif
        ret = 0;
        goto skip;
    }
    else {
#if 0
        Trace("no refered , *purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }

skip:
    return ret;

}



int
skcache_entry_purge_check(void *pp, void *vv)
{
    skcache_entry_t *p = (skcache_entry_t*)pp;
    bucket_t *bk = (bucket_t*)vv;
    int ret = 0;
    time_t now;
    int elap, idle;
    int erase;

    erase = 0;

    now = time(NULL);
    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
#if 0
        Trace("pc: ignore issue_time\n");
#endif
        elap = skcache_mem_keep_max;
    }
    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
        idle = skcache_mem_keep_cur/2;
    }

    if(p->ref>0) {
        erase = 0;
    }
    else
    if(p->expire>0 && p->expire >= now) {
        erase = 1;
    }
    else
    if(idle >= skcache_mem_keep_cur) {
        erase = 1;
    }

#if 0
    Trace("  check %p, ref %d, freq %d, idle %4d/%d, elap %5d: ",
        p, p->ref, p->count, idle, skcache_mem_keep_cur, elap);
#endif

    if(erase>0) {
#if 0
        Trace("*purge*\n");
#endif
        skcache_entry_purge(bk, p);
        skcache_push_idle_entry(bk,  p);
        p = NULL;

        ret = 1;
    }
    else {
#if 0
        Trace("not purge\n");
#endif
        ret = 0;
    }

skip:
    return ret;

}


#ifdef NO_VS
#ifdef VS
#undef VS
#endif
#endif



/*
 * return code is describe in comment in the routin.
 */
int
skcache_dicision_w_th(bucket_t *bk, skcache_entry_t *p, int force,
    time_t now, time_t min, time_t th, time_t max)
{
    int elap, idle;
    int action;
    int chk;

#if 0
    Trace(
"_dicision_w_th: bk# %d, p 0x%x, force %d // min,thre,max %d,%d,%d\n",
        bk->bid, p, force, min, th, max);
#endif

    action = 0;

    if(p->issue_time > 0 ) {
        elap = now - p->issue_time;
    }
    else {
        elap = -1;
    }

    if(p->oid == 0 ) {
        if(p->owner) {
#if 0
            Trace(
"_dc: temporary  p %p %d oid# %d, ref %d, count %d, size %6ld, owner %p\n",
    p, p->issue_time, p->oid, p->ref, p->count, (long)p->totallength, p->owner);
#endif
            action = 0;
            goto skip;
        }
        else {
#if 0
            Trace(
"_dc: losted tmp p %p %d oid# %d, ref %d, count %d, size %6ld, owner %p\n",
    p, p->issue_time, p->oid, p->ref, p->count, (long)p->totallength, p->owner);
#endif
            action = 2;
            goto skip;
        }
    }


#if 0
    else {
#if 1
        Error(
"_dc: ignore issue_time %d <addr %p>, oid# %d, ref %d, count %d, size %ld, owner %p\n\tname '%s'",
            p->issue_time, p, p->oid, p->ref, p->count, (long)p->totallength,
            p->owner, p->name);

#if 0
        if(p->owner) {
        Error("_dc: ignore issue_time <addr %p>, oid# %d, name '%s', owner %p sid# %d,",
            p, p->oid, p->name, p->owner, ((sess_t*)p->owner)->sid);
        }
        else {
        Error("_dc: ignore issue_time <addr %p>, oid# %d, name '%s', no owner",
            p, p->oid, p->name);
        }
#endif
#endif
        elap = -1;

#if 0
        action = 2;
        goto skip;
#endif
    }

#endif

    if(p->hot > 0 ) {
        idle = now - p->hot;
    }
    else {
#if 0
        Trace("empty hot\n");
#endif
#if 0
        idle = MIN(th/2, elap);
#endif
        idle = -1;
    }

#ifdef VS
    if(p->m_recorded) {
    Trace(" %p:: %8d %7ld %c 0x%08x %3d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : '-',
        p->m_recorded, p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
    else {
    Trace(" %p:: %8d %7ld %c %5s %8d, c %d, f %2d, idle %4d, elap %5d: ",
        p, p->oid, (long)p->totallength,
        (p->ref > 0) ? 'R' : ' ',
        "-", p->d_recorded, p->cracked, p->count,
        idle, elap);
    }
#endif

#if 0
    Trace("lmd %d now %d diff %d ", p->lmd, now, now - p->lmd);
#endif

    /* action
        3   swapout + purge
        2   purge
        1   swapout
        0   no action
     */

    if(p->ref>0) {
#ifdef VS
Trace("refered ");
#endif
        action = 0;
    }
    else
    if(elap<0) {
#ifdef VS
Trace("junk    ");
#endif
        action = 2;
    }
    else
    if(p->cracked) {
#ifdef VS
Trace("cracked ");
#endif
        action = 2;
    }
    else
    if(p->d_recorded>=0) {
        if(p->m_recorded==NULL) {
#ifdef VS
Trace("already ");
#endif
            /* already swapouted */
            action = 0;
        }
        else {
#ifdef VS
Trace("mine-01 ");
#endif
            /* XXX */
            action = 1;
        }
    }
    else
    if(p->m_recorded==NULL) {
#ifdef VS
Trace("empty-b ");
#endif
        action = 2;
    }
    else
    if(p->totallength<=(off_t)0) {
#ifdef VS
Trace("empty-c ");
#endif
        action = 2;
    }
    else
    if(elap > max) {
#ifdef VS
Trace("expire1 ");
#endif
        action = 2;
    }
    else
    if(p->expire>0 && p->expire <= now) {
#ifdef VS
Trace("expire2 ");
#endif
        action = 2;
    }
#if 0
    else
    if(p->lmd>0 && now - p->lmd > max) {
#ifdef VS
Trace("expire3 ");
#endif
        action = 2;
    }
#endif
    else {

#if 0
        /*
         * XXX
         * randomize forward swapout feature
         */
        if(idle<=min) {
            /* nothing */
        }
        else {
            idle += rand() % th;
        }
#endif

        if(force) {
            if(idle < 0) {
#ifdef VS
    Trace("neg,    ");
#endif
                action = 2;
            }
            else {
#ifdef VS
    Trace("shift   ");
#endif
                action = 3;
            }
        }
        else {
            if(idle < 0) {
#ifdef VS
    Trace("neg,    ");
#endif
                action = 2;
            }
            else
            if(idle >= th) {
#ifdef VS
    Trace("shift   ");
#endif
                action = 3;
            }
            else {
#ifdef VS
    Trace("keep    ");
#endif
            }
        }
    }

skip:
    return action;

}

int
skcache_dicision(bucket_t *bk, skcache_entry_t *p, int force)
{
    time_t now;

    now = time(NULL);

    return skcache_dicision_w_th(bk, p, force,
        now,
        skcache_mem_keep_min, skcache_mem_keep_cur, skcache_entry_keep_cur);
}




int
rm_fr(char *base)
{
    char tmp[L_PATH];
    dirent_t    *dp;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    struct stat sbuf;
    int chk;

    if(base==NULL)
        return -1;

#if 0
Trace("rm_fr: called '%s'\n", base);
#endif

    chk = rmdir(base);
    if(chk) {
        if(errno==ENOTEMPTY) {

            dirp = opendir(base);
            if(dirp) {
                while(1) {
                    chk = readdir_r(dirp, ent, &dp);
                    if(dp==NULL) {
                        break;
                    }

                    if(dp->d_name[0]=='.' && dp->d_name[1]=='\0') {
                        continue;
                    }
                    if(dp->d_name[0]=='.'
                        && dp->d_name[1]=='.' && dp->d_name[2]=='\0') {
                        continue;
                    }
                    /*
                    strcpy(tmp, base);
                    strcat(tmp, "/");
                    strcat(tmp, dp->d_name);
                    */
                    pathncat(tmp, L_PATH, base, dp->d_name);


                    chk = stat(tmp, &sbuf);
                    if(chk) {
                        Error("rm_fr: fail stat() '%s' (%d)",
                            tmp, errno);
                        continue;
                    }

                    if(S_ISREG(sbuf.st_mode)) {
                        unlink(tmp);
                    }
                    if(S_ISDIR(sbuf.st_mode)) {
                        rm_fr(tmp);
                    }
                }
                closedir(dirp);
            }

            chk = rmdir(base);

            if(chk) {
                Error("rm_fr: fail rmdir() '%s' after sweep (%d)",
                    base, errno);
            }

        }
    }

    return 0;
}





extern pthread_mutex_t stat_lock;
extern volatile int compaction_count;

typedef struct {
    int no;
    int type;   /* 0 single, 1 packed */
    int pos;    /* position */
    int act;
    time_t date;
    off_t sz;
} tup;

int
comp_tup(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->act == y->act) {
        if(x->date == y->date) {
            return x->sz - y->sz;           /* sz smaller */
        }
        else {
            return y->date - x->date;       /* date grater */
        }
    }
    else {
        return x->act - y->act;             /* act smaller */
    }
}

int
comp_tup_order(const void* xp, const void* yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    if(x->type == y->type) {
        return x->pos - y->pos;             /* y-x smaller */
    }
    else {
        return y->type- x->type;            /* y-x greater */
    }
}

int
comp_tup_time(void*xp,void*yp)
{
    tup *x = (tup*)xp;
    tup *y = (tup*)yp;

    return x->date - y->date;
}

int
show_tup(char *msg, int num, tup *base)
{
    int i;
    tup *cur;
    off_t lsum, osum;

    Trace("show_tup: %s\n", msg);

    lsum = osum = (off_t) 0L;
    cur = base;
    for(i=0;i<num;i++) {
        if(cur->act==ACTION_NOP || cur->act==ACTION_SKIP) {
        lsum += cur->sz;
        Trace(" %12d: %c %4d : %d %10d %12lld %12lld\n",
            cur->no, cur->type ? 'P' : 's', cur->pos,
            cur->act, cur->date, (long long)cur->sz, (long long)lsum);
        }
        else {
        osum += cur->sz;
        Trace(" %12d: %c %4d : %d %10d %12lld\n",
            cur->no, cur->type ? 'P' : 's', cur->pos,
            cur->act, cur->date, (long long)cur->sz);
        }
        cur++;
    }
}


int 
check_target(skcache_entry_t *en, time_t reftime, int *xaction)
{
    int xh1, xh2, xh3, xh4;
    int igh;

#if 0
Trace("check_target: en %p, MAGIC %x %x, oid# %d, *xaction %d\n",
    en, (unsigned char)en->magic[0], (unsigned char)en->magic[1],
    en->oid, *xaction);
#endif

    if(en->magic[0]!=MAGIC_0 || en->magic[1]!=MAGIC_1) {
        *xaction = ACTION_UNLINK;
#ifdef CHECK_MAGIC
Error("check_target: oid# %d, wrong magic %02x %02x / %02x %02x",
        en->oid, 
        en->magic[0], en->magic[1], MAGIC_0, MAGIC_1);
#endif
        en->issue_time = -1;
        goto out;
    }

    if(en->oid == 0) {
#if 0
Trace("oid# %d, *xaction %d -> 0\n", en->oid, *xaction);
#endif
        *xaction = ACTION_SKIP;
        goto out;
    }

    calc_hash(en->name, &xh1, &xh2, &xh3, &xh4);

    if(en->h1!=xh1 || en->h2!=xh2 || en->h3!=xh3 || en->h4!=xh4) {
        *xaction = ACTION_UNLINK;
        igh++;
#if 0
Trace("\
check_target: ignore hash, file '%s'\n\
    stored %03d:%03d:%03d:%03d %s\n\
    solved %03d:%03d:%03d:%03d same\n",
    name,
    en->h1, en->h2, en->h3, en->h4, en->name,
    xh1, xh2, xh3, xh4);
#endif
#if 0
Trace("check_target: ignore hash, stored %03d:%03d:%03d:%03d, we guess %03d:%03d:%03d:%03d\n",
        en->h1, en->h2, en->h3, en->h4, xh1, xh2, xh3, xh4);
#endif
        en->issue_time = -1;
        goto out;
    }


    if(en->issue_time<=0) {
        *xaction = ACTION_PURGE;
    }
    else {
#if 1
        /*XXX*/
        if(en->issue_time<min_date) {
            *xaction = ACTION_UNLINK;
        }
        else
        if(en->issue_time>reftime) {
            *xaction = ACTION_UNLINK;
        }
        else
#endif
        if(reftime - en->issue_time > skcache_entry_keep_cur) {
            *xaction = ACTION_PURGE;
        }
    }

    if(!*xaction) {
        if(en->lmd<=0) {
            *xaction = ACTION_PURGE;
        }
#if 0
        if(en->lmd>0 && reftime - en->lmd > skcache_entry_keep_cur) {
            *xaction = ACTION_PURGE;
        }
#endif

        if(en->expire>0 && reftime - en->expire > 0) {
            *xaction = ACTION_PURGE;
        }
    }


out:
    return *xaction;
}




/*
 * convert num-num -> num
 */
int name2num(char *nm)
{
    int r1, r2, r3, r4;
    int ret;
    register char *p, *q;
    register int i;
    char tmp[L_PATH];
    
    p = nm;

    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r3 = atoi(tmp);
    
    q = tmp;
    i = 0;
    while(*p && (*p>='0' && *p<='9') && i < L_PATH) {
        *q++ = *p++;
        i++;
    }
    p++;
    *q = '\0';
    r4 = atoi(tmp);

    ret = r3*radix_h4 + r4;

#if 0
    Trace("name2num '%s' -> %d\n", nm, ret);
#endif

    return ret;
}


int
skcache_check_dir(int d1, int d2, int d3, int force)
{
    char basename[L_PATH];
    char dirname[L_PATH];
    char tmp[L_PATH];
    char *namechk;
    dirent_t    *elem;
    DIR         *dirp;
#ifdef PC_NAME_MAX
    dirent_t ent[PC_NAME_MAX+1];
#else
#ifdef _PC_NAME_MAX
    dirent_t ent[_PC_NAME_MAX+1];
#else
    dirent_t ent[128];
#endif
#endif
    char name[L_PATH];
    int dfd;
    char buf[L_BUF];
    int n;
    int i;
    int j, k;
    int nr;
    int chk;
    off_t sum;
    off_t fsize;
    int lcount, dcount, ecount;
    off_t lsize, dsize, esize;
    int strange = 0;
    int action;
    int life;
    time_t reftime;
    time_t now;
    int nfile;

    int lastpos;
    off_t dst;

    tup rec[TUP_MAX];

    struct timeval begin, end;

    bucket_t *bk;
    skcache_entry_t en;
    skcache_entry_t chunk[NINDEX_DREAD_MAX];
    int c;
    int er=0;

    int igh=0;
    int nw;

    u_int32_t sig[DSIG_ELEM_WIDTH];

    int sig_pos = 0;

    int has_index;
    int has_packed;

    int dmark_count;



    gettimeofday(&begin, NULL);
    reftime = begin.tv_sec;

    now = time(NULL);
    sig_pos = DSIG_POS(d1, d2, d3);


#if VD
Trace("check_dir: [%4d,%4d,%4d] dir# %d, swaper# %d, time %d force? %d\n",
   d1, d2, d3, sig_pos, d1%MAX(radix_disk, 1), reftime, force);
#endif


    lcount  = 0;
    dcount  = 0;
    lsize   = 0;
    dsize   = 0;
    ecount  = 0;
    esize   = 0;


    if(d1<0||d1>=RADIX_H1_MAX
        || d2<0||d2>=RADIX_H2_MAX
            || d3<0||d3>=RADIX_H3_MAX) {
        Error("check_dir: ignore directory position %d,%d,%d", d1, d2, d3);
        goto out;
    }

    if(d1>=radix_h1 || d2>=radix_h2 || d3>=radix_h3) {
        if(force) {
            Trace("check_dir: out of current directory range <%d,%d,%d>, try clear\n",
                d1, d2, d3);

            namechk = assign_dirname_bynumber(dirname, d1, d2, -1);
            if(namechk) {
                Trace("check_dir: clear dir. recursive %s\n", dirname);
                rm_fr(dirname);
            }
        }
        else {
            Trace("check_dir: out of current directory range <%d,%d,%d>, no change\n",
                d1, d2, d3);
        }

        goto out;
    }


    dsig_allclear(sig);

    bk = &skcache_bucket_base[d1];

#ifdef VD2
    if(dir_sz[sig_pos]>=dir_size_hwater) {
    Trace("  check_dir: dir# %d, dir_sz %lld %.2fM %.1f%%, sum %lld %.1f%%, lsize %lld %.1f%%\n",
        sig_pos,
        (long long)dir_sz[sig_pos], fdM(dir_sz[sig_pos]),
            (double)100.0*dir_sz[sig_pos]/dir_size_max,
        (long long)sum, (double)100.0*sum/dir_size_max,
        (long long)lsize, (double)100.0*lsize/dir_size_max);
    }
#endif


disk_operations:


    n = 0;
    sum = (off_t)0L;

    namechk = assign_dirname_bynumber(dirname, d1, d2, d3);

#ifdef VD3
    Trace("  target dir# %d, path '%s'\n", sig_pos, dirname);
#endif


    /***
     *** scan directories
     ***/

    c = 0;
try_opendir:
    c++;
    dirp = opendir(dirname);
    if(dirp==NULL) {
        if(errno==ENOENT && c<DISK_RETRY) {
            if(force) {
#ifdef VD
                Trace("check_dir: not found dir# %d, '%s', build that.\n",
                    sig_pos, dirname);
#endif
                build_dir_bynumber(d1, d2, d3);

                MUTEX_LOCK(&dir_mutex);
                checkdir_dir_count++;

                DSIG_ALLCLEAR(sig_pos);
                dir_sz[sig_pos]     = (off_t)0L;
                dir_count[sig_pos]  = 0L;

                MUTEX_UNLOCK(&dir_mutex);

                msleep(10);     /* XXX */
            }

            goto fail_opendir;
        }
        if(errno==EINTR && c<DISK_RETRY) {
            goto try_opendir;
        }

        Error("check_dir: fail opendir() '%s' %d (%d)", dirname, c, errno);

        goto fail_opendir;
    }



    /******
     ****** sweep single files from directory
     ******
     ******
     ******/

    has_index = 0;
    has_packed = 0;

    i = 0;
    while(1) {
        chk = readdir_r(dirp, ent, &elem);
        if(chk<0) {
            if(errno==EINTR) {
                continue;
            }
            Error("fail readdir_r() for '%s' (%d)", dirname, errno);
            break;
        }
        if(elem==NULL) {
            break;
        }

        if(elem->d_name[0]=='.' && elem->d_name[1]=='\0') {
            continue;
        }
        if(elem->d_name[0]=='.'
            && elem->d_name[1]=='.' && elem->d_name[2]=='\0') {
            continue;
        }

        pathncat(name, L_PATH, dirname, elem->d_name);
#ifdef VD
        Trace("    %6d %4d : %-14s %s\n", sig_pos, i, elem->d_name, name);
#endif

        if(elem->d_name[0]==INDEX_FILENAME0
            && strcmp(elem->d_name, INDEX_FILENAME)==0) {
            has_index = 1;
            continue;
        }
        if(elem->d_name[0]==PACKED_FILENAME0
            && strcmp(elem->d_name, PACKED_FILENAME)==0) {
            has_packed = 1;
            continue;
        }


        i++;

        life = -1;
        action = ACTION_NOP;

        c = 0;
try_open:
        c++;
        dfd = open(name, O_RDONLY);
        if(dfd<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_open;
            }
            Error("check_dir: fail open() '%s' %d (%d)",
                name, c, errno);
            action = ACTION_PURGE;
            goto next;
        }

        /* count file size by seek end
            and seek first to read control block */
        c = 0;
try_lseek:
        c++;
        fsize = lseek(dfd, 0, SEEK_END);
        if(fsize<(off_t)0) {
            if((errno==EINTR || errno==EINPROGRESS) && c<DISK_RETRY) {
                msleep(3);
                goto try_lseek;
            }
            Error("check_dir: fail lseek() '%s' %d (%d)",
                name, c, errno);
            action = ACTION_PURGE;
            goto done_read_cntl;
        }
        if(fsize>(off_t)0) {
            lseek(dfd, 0, SEEK_SET);
        }

        sum += fsize;

        c = 0;
try_read_cntl:
        c++;
        nr = read(dfd, &en, sizeof(skcache_entry_t));
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read_cntl;
            }
            Error("check_dir: file# %d, fail read() '%s' %d (%d)",
                n, name, c, errno);
            action = ACTION_PURGE;
            goto done_read_cntl;
        }

        if(nr<sizeof(skcache_entry_t)) {
            action = ACTION_UNLINK;
            Error("check_dir: ignore file# %d, read control short than control %d/%d",
                n, nr, sizeof(skcache_entry_t));
#if 1
#endif
            en.issue_time = -1;
            goto done_read_cntl;
        }

        
        check_target(&en, reftime, &action);

done_read_cntl:
        (void) 1;

read_done:

        close(dfd);

next:

#ifdef VD3
        Trace(" f %4d %4d %4d %d %d %8d %9ld oid# %d, %s\n",
            i, n, atoi(elem->d_name), action,
            en.issue_time, reftime - en.issue_time, (long)fsize, en.oid, name);
#endif

        if(n>=TUP_MAX) {
            er++;
            if(er==1) {
                Error("dir '%s', tup overflow read %d th, store %d th in %d",
                    dirname, i, n, TUP_MAX);
            }
        }
        else {
            rec[n].no = name2num(elem->d_name);

/*
Trace("n %d %d <- %s\n", n, rec[n].no, elem->d_name);
*/
            rec[n].type = 0;
            rec[n].pos  = 0;
            rec[n].act = action;
            rec[n].date = en.issue_time;
            rec[n].sz = fsize;
            n++;
        }
    }

    closedir(dirp);

    nfile = i;


    /******
     ******
     ****** sweep objects in packed (index) file
     ******
     ******/

    if(!has_index) {
#ifdef VD
        Trace("not have index file, skip\n");
#endif
        dfd = -1;
        goto skip_packedfile;
    }


    if(!has_packed) {
        goto skip_packedfile;
    }
    
/*
    pathncat(name, L_PATH, dirname, PACKED_FILENAME);
    Trace("packed file name '%s'\n", name);
*/
    pathncat(name, L_PATH, dirname, INDEX_FILENAME);


    en.h1 = d1;
    en.h2 = d2;
    en.h3 = 0;
    en.h4 = 0;

    chk = ifd_open(&dfd, bk, &en);
    if(chk) {
        goto skip_packedfile;
    }

    c = 0;
try_lseek2:
    c++;
    fsize = lseek(dfd, 0, SEEK_END);
    if(fsize<(off_t)0) {
        if((errno==EINTR || errno==EINPROGRESS) && c<DISK_RETRY) {
            msleep(3);
            goto try_lseek2;
        }
        Error("check_dir: fail packed lseek() '%s' %d (%d)",
            name, c, errno);
        goto skip_packedfile;
    }
    if(fsize>(off_t)0) {
        lseek(dfd, 0, SEEK_SET);
        sum += fsize;
    }


    for(j=0;j<OFILE_SLOT;j += nindex_dread ) {
        c = 0;
try_read_cntl2:
        c++;
        nr = read(dfd, chunk, sizeof(skcache_entry_t)*nindex_dread);
        if(nr<0) {
            if(errno==EINTR && c<DISK_RETRY) {
                goto try_read_cntl2;
            }
            Error("check_dir: file# %d, fail read() '%s' %d (%d)",
                n, name, c, errno);
            goto done_read_cntl2;
        }
        if(nr<sizeof(skcache_entry_t)) {
            action = ACTION_PURGE;
            Error("check_dir: ignore target# %d, pos %d, read control in packed short than control %d/%d",
                n, j, nr, sizeof(skcache_entry_t));
            goto fill_rest;
        }

        for(k=0;k<nindex_dread;k++) {
            action = ACTION_NOP;
            en = chunk[k];

            check_target(&en, reftime, &action);

#ifdef VD3
            Trace(" p %4d %4d %4d %d %9d %9d %9ld oid# %d, -\n",
                j+k, n, j+k, action,
            en.issue_time, reftime - en.issue_time, (long)fsize, en.oid);
#endif
            if(n>=TUP_MAX) {
                er++;
                if(er==1) {
                    Error(
                        "dir '%s', tup overflow read %d th, store %d th in %d",
                        dirname, i, n, TUP_MAX);
                }
                break;
            }

            rec[n].no = en.h4; 

/*
Trace("n %d %d <- %s\n", n, rec[n].no, elem->d_name);
*/
            rec[n].type = 1;    /* packed */
            rec[n].pos  = j+k;
            rec[n].act  = action;
            rec[n].date = en.issue_time;
            rec[n].sz   = OFILE_UNIT; /* XXX */
            n++;
        }
    }

    goto done_read_cntl2;

fill_rest:
    Trace("  dir# %d, fill rest [%d..%d) of packed file\n",
        sig_pos, j, OFILE_SLOT);

    for(;j<OFILE_SLOT;j++) {
        rec[n].no = -1;
/*
Trace("n %d %d <- %s\n", n, rec[n].no, elem->d_name);
*/
        rec[n].type = 1;    /* packed */
        rec[n].pos  = j;
        rec[n].act  = 2;
        rec[n].date = 0;
        rec[n].sz   = OFILE_UNIT; /* XXX */
        n++;

    }






done_read_cntl2:
    (void)1;

    
skip_packedfile:



#ifdef VD
    Trace("  dir# %d, nfile %d, record %d/%d, \n",
        sig_pos, nfile, n, TUP_MAX);
#endif

    if(nfile<=0 && has_index==0) {
        goto report_phase;
    }
#ifdef VD3
    show_tup("after  dir.sweeping", n, rec);
#endif

#ifdef VD3
    Trace("  sweeped file size, sum %lld\n", (long long) sum);
#endif


    /***
     *** main body of compaction ***/

#ifdef VD3
    show_tup("before sort1", n, rec);
#endif


    /* 
     * sort by action, date and size
     *      black and gray objects will locate to top.
     *
     */
    qsort(rec, n, sizeof(tup), comp_tup);
#ifdef VD2
    show_tup("after  sort1", n, rec);
#endif

#ifdef VD2
    Trace("  lwater %lld\n", (long long)dir_size_lwater);
#endif

    for(i=0;i<n;i++) {
        int xh1, xh2, xh3, xh4;

        xh4 = rec[i].no;
#if 0
        Trace("%4d: no# %7d act %2d date %d size %12lld\n",
            i, rec[i].no, rec[i].act, rec[i].date, (long long)rec[i].sz);
#endif

        if(lsize + rec[i].sz > dir_size_lwater) {
#ifdef VD2
            Trace("  change to PURGE from idx %4d pos %4d\n", i, rec[i].pos);
#endif
            break;
        }

        switch(rec[i].act) {
        case ACTION_NOP:
            /* mark with xh3 and xh4 */
            DSIG_TMP_SET(sig, xh4/DSIG_VALUE_RES);
            lcount++;
            lsize += rec[i].sz;
            break;

        case ACTION_PURGE:
        case ACTION_UNLINK:
            dcount++;
            dsize += rec[i].sz;
            break;

        case ACTION_SKIP:
            ecount++;
            esize += rec[i].sz;
            break;

        default:
            break;

        }
    }
    for(;i<n;i++) {
        rec[i].act = ACTION_PURGE;
        dcount++;
        dsize += rec[i].sz;
    }

#ifdef VD
        Trace(
"\tthen, l %9lld d %9lld e %9lld : free %12lld %.2f%%\n",
            (long long)lsize, (long long)dsize, (long long)esize,
            (long long) (dir_size_max - lsize),
            (double)(dir_size_max - lsize)*100.0/dir_size_max
             );
#endif

    /*
     * to optimize disk access, sort them with location order
     *
     */
    qsort(rec, n, sizeof(tup), comp_tup_order);

#ifdef VD3
    show_tup("after  sort2", n, rec);
#endif

    for(i=0;i<n;i++) {
        int xh1, xh2, xh3, xh4;

        if(rec[i].type==1) {
            break;
        }
        if(rec[i].type==0) {
        }

        xh4 = rec[i].no;
        sprintf(name, "%s/%d", dirname, xh4);

#ifdef VD4
Trace("i %d no %d xh4 %d -> %s\n", i, rec[i].no, xh4, name);
#endif

        switch(rec[i].act) {
        case ACTION_PURGE:
            c = 0;
try_unlink1:
            c++;
            chk = unlink(name);
            if(chk) {
                if(errno==EINTR) {
                    goto try_unlink1;
                }
                if(errno==EISDIR) {
                    rm_fr(name);
                }
                else {
                    Error("check_dir: 1 - fail unlink() %s %d (%d)",
                        name, c, errno);
                }
            }
            break;

        case ACTION_UNLINK:
            c = 0;
try_unlink2:
            c++;
            chk = unlink(name);
            if(chk) {
                if(errno==EINTR) {
                    goto try_unlink2;
                }
                Error("check_dir: 2 - fail unlink() %s %d (%d)",
                    name, c, errno);
            }

            break;
        }
    }

    en.magic[0] = MAGIC_0;
    en.magic[1] = MAGIC_1;
    en.h4   = -73;
    en.oid  = 0;
    en.issue_time = -1;

    dmark_count = 0;
    dst = lseek(dfd, 0, SEEK_SET);
    
    lastpos = -1;
    for(;i<n;i++) {
        int xh1, xh2, xh3, xh4;

        xh4 = rec[i].no;

        switch(rec[i].act) {
        case ACTION_PURGE:
        case ACTION_UNLINK: 
            if(lastpos+1 != rec[i].pos) {
#ifdef VD2
    Trace(" %4d last %4d %12lldB, cur %4d %12lldB, diff %4d %12lldB\n",
        i, 
        lastpos,    (long long)lastpos* sizeof(skcache_entry_t),
        rec[i].pos, (long long)rec[i].pos * sizeof(skcache_entry_t),
        rec[i].pos - lastpos - 1,
            (long long)(rec[i].pos - lastpos -1) * sizeof(skcache_entry_t));
#endif

                c = 0;
try_lseek3:
                c++;
                dst = lseek(dfd,
                        (rec[i].pos - lastpos -1) * sizeof(skcache_entry_t),
                        SEEK_CUR);
                if(dst<(off_t)0) {
                    if(errno==EINTR && c<DISK_RETRY) {
                        goto try_lseek3;
                    }
Error("check_dir: cannot seek abs. %lld on index file fd# %d :%d (%d)",
            (long long)rec[i].pos* sizeof(skcache_entry_t), dfd, c, errno);

                    goto report_phase;
                }
            }

            c = 0;
try_write:
            c++;
            nw = write(dfd, &en, sizeof(skcache_entry_t));
            if(nw<0) {
                if(errno==EINTR && c<DISK_RETRY) {
                    goto try_write;
                }
                Error("check_dir: fail deletation marking dfd# %d, (%d)",
                    dfd, errno);
            }
            dmark_count++;

            lastpos = rec[i].pos;
            break;
        }
    }

#ifdef VD
    Trace("  write deletation mark %d\n", dmark_count);
#endif

report_phase:

    now = time(NULL);

#ifdef VD
    Trace(" dd %6d %-24s %4d sum %8ld + %d\
/ live %4d %8ld / dead %4d %8ld\
/ untouch %4d %8ld / %2d\n",
        sig_pos, dirname, n, (long)sum, strange,
        lcount, (long)lsize, dcount, (long)dsize, ecount, (long)esize,
        now - reftime
        );
#endif

    MUTEX_LOCK(&stat_lock);

    checkdir_dir_count++;
    checkdir_file_read_count    += nfile;
    checkdir_file_handle_count  += n;
    checkdir_file_left_count    += lcount;
    checkdir_file_delete_count  += dcount;
    checkdir_file_erase_count   += ecount;

    check_dir_ignore_hash_count = igh;

    if(dcount>0 && dsize>0) {
        swap_count  -= dcount;
        swap_size   -= dsize;
    }

    MUTEX_UNLOCK(&stat_lock);

    if(force) {
        MUTEX_LOCK(&dir_mutex);

        memcpy((void*)&dir_sig[sig_pos*DSIG_ELEM_WIDTH],
            (void*)&sig, DSIG_ELEM_WIDTH*sizeof(u_int32_t));

        dir_sz[sig_pos]     = lsize;
        dir_count[sig_pos]  = lcount;
        dir_time[sig_pos]   = now;

        MUTEX_UNLOCK(&dir_mutex);
    }

fail_opendir:

out:

    gettimeofday(&end, NULL);
    sim_stat_add_value(&checkdir_time, KOTETU_SUB_TIME(end, begin));
    sim_stat_add_value(&checkdir_time_short, KOTETU_SUB_TIME(end, begin));

#ifdef TRACE_CHKDIR
    Trace("  done chkdir dir# %d, %-17s left %4df %8ldB %7.2fKB  th %08x\n",
            sig_pos, dirname, lcount, (long)lsize, fdK(lsize), pthread_self());
#endif
#if 0
#endif

end:
    return 0;
}



volatile int dsweeper_seq   = 0;
volatile int dsweeper_count = 0;
volatile int dsweeper_phase = -1;
volatile int dsweeper_pos   = -1;
volatile int dsweeper_full_mode = 0;

/* how many tics are required to work sweeper */
#ifndef SWEEP_CYCLE
#define SWEEP_CYCLE (1)
#endif
#ifndef SWEEP_BOOT_CYCLE
#define SWEEP_BOOT_CYCLE (1)
#endif

/* how many direcotry are sweeped per chance */
#ifndef SWEEP_NTRY
#define SWEEP_NTRY  (1)
#endif


volatile int dsweep_boot_cycle = SWEEP_BOOT_CYCLE;
volatile int dsweep_cycle = SWEEP_CYCLE;



/* for two level */
int
skcache_dir_sweeper(void)
{
    int f;
    int d1, d2, d3;
    int v;
    int chk;
    int p;
    int nd1;
    int pos;

    MUTEX_LOCK(&stat_lock);
    if(dsweeper_phase<0) {
        dsweeper_phase = 0;
        dsweeper_pos = rand() % dir_num;
    }
    p = dsweeper_phase;
    f = dsweeper_seq;
    dsweeper_seq =  (dsweeper_seq+1) % 1000000;

#if 0
    if(dsweeper_count%dir_num==0) {
        Trace("first dir\n");
    }
#endif
    dsweeper_count++;

    v = dsweeper_pos;

    if(dsweeper_pos>=dir_num-1) {
        dsweeper_phase++;
        dsweeper_pos = 0;
    }
    else {
        dsweeper_pos++;
    }
    MUTEX_UNLOCK(&stat_lock);

    if(f<0) {
        f = v;
    }

    d1  = v%radix_h1;
    d2  = v/radix_h1;
    d3  = 0;

    pos = DSIG_POS(d1, d2, d3);

#if 0
    Trace("  sweeper th %08x %6d %6d -> dir# %d, %03d %03d %03d\n",
        pthread_self(), p, v, pos, d1, d2, d3);
#endif
    MUTEX_LOCK(&dir_mutex);
    if(dir_lock[pos]>0) {
        pos = -1;
    }
    MUTEX_UNLOCK(&dir_mutex);

    if(pos>0) {
        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = pthread_self();
        MUTEX_UNLOCK(&dir_mutex);

        chk = skcache_check_dir(d1, d2, d3, 1);

        MUTEX_LOCK(&dir_mutex);
        dir_lock[pos] = 0;
        MUTEX_UNLOCK(&dir_mutex);
    }

done:

    return 0;
}

int
skcache_calc_param(long long want,
    int *xbk_num, int *xsk_level, int *xffl_size, double r,
    int *xsk_len, int *xffl_num)
{
    long long bksize;
    long a, b, c;

#if 0
    Trace("skcache_calc_param: want %lld\n", want);
#endif

#if 0
    Trace("    specified xbk_num %d, xsk_level %d, xffl_size %d r %.2f\n",
        *xbk_num, *xsk_level, *xffl_size, r);
#endif

    if(*xffl_size<0) {
        *xffl_size = 1460;
    }

    bksize = want/(*xbk_num) - sizeof(skcache_entry_t);

    a = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    b = sizeof(skcache_entry_t);
    c = sizeof(skcache_frag_t)+ *xffl_size;

#if 0
    Trace("    bucket size %lld\n", bksize);
    Trace("    a %ld\n", a);
    Trace("    b %ld\n", b);
    Trace("    c %ld\n", c);
#endif

    *xsk_len = (long)
        (bksize - b) / (a + b + (long)(r*c));
    *xffl_num = (long)(r * *xsk_len);

#if 0
    Trace("    solved    xsk_len %d, xffl_num %d\n",
        *xsk_len, *xffl_num);
#endif


    return *xsk_len;
}


long long
skcache_calc_size(
    int *xbk_num, int *xsk_level, int *xsk_len, int *xffl_len, int *xffl_size)
{
    long long r;
    int esk;
    int tsk;
    int ten;
    int tfl;
    long long ebk;
    long long tbk;

#if 0
    Trace("skcache_calc_size:\n");
    Trace("  bucket:\n");
#endif

    r = (((sizeof(void*)+sizeof(skiplist_cell_t**)+sizeof(skiplist_cell_t*)*(*xsk_level+1)) + sizeof(skcache_entry_t))* (*xsk_len)
        + sizeof(skcache_entry_t)
        + (sizeof(skcache_frag_t) + *xffl_size) * (*xffl_len))*(*xbk_num);

    esk = sizeof(void*)+sizeof(skiplist_cell_t**)
            +sizeof(skiplist_cell_t*)*(*xsk_level+1);
    tsk = esk * *xsk_len;
#if 0
    Trace("    skiplist:\n");
    Trace("      level %d\n", *xsk_level);
    Trace("      -> element size of skiplist %ld\n", esk);
    Trace("      length of skiplist %d\n", *xsk_len);
#endif

    ten = sizeof(skcache_entry_t)*(*xsk_len+1);
    tfl = (sizeof(skcache_frag_t)+ *xffl_size)*(*xffl_len);
    ebk = (long long)(sizeof(bucket_t)+tsk+ten+tfl);
#if 0
    Trace("    -> size of skiplist            %12ld %4ldM+%07ld %7.2f%%\n",
        tsk, (long)dM(tsk), (long)mM(tsk), ((double)tsk*100)/ebk);
    Trace("    size of free-list for entities %12ld %4ldM+%07ld %7.2f%%\n",
        ten, (long)dM(ten), (long)mM(ten), ((double)ten*100)/ebk);
    Trace("    size of fragment free-list     %12ld %4ldM+%07ld %7.2f%%\n",
        tfl, (long)dM(tfl), (long)mM(tfl), ((double)tfl*100)/ebk);
    Trace("  -> size of bucket                %12lld %4ldM+%07ld\n",
        ebk, (long)dM(ebk), (long)mM(ebk));
#endif


    tbk = ebk * *xbk_num;

#if 0
    Trace("  number of bucket %ld\n", *xbk_num);
    Trace("-> total size of all buckets       %12lld %4ldM+%07ld\n",
        tbk, (long)dM(tbk), (long)mM(tbk));
    Trace("-> total number of entries         %12ld\n", *xsk_len * *xbk_num);
    Trace("-> total number of frags           %12ld\n", *xffl_len * *xbk_num);
#endif

    return r;
}




int check_bkinfo();


skcache_entry_t*
skcache_pull_new_entry(bucket_t *bk)
{
    int chk;
    skcache_entry_t* newone;

#if 0
#ifndef NDEBUG
    Trace("new_entry: bk# %d, %p, use %d, free %d, cup.num %d\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num);
#endif
#endif

    if(bk->entry_freelist == NULL) { 
        Error("new_entry: no free item");
        return NULL;
    }

    newone = bk->entry_freelist;
    bk->entry_freelist = newone->next;
    bk->entry_use++;
    bk->entry_free--;

    memset(newone, 0, sizeof(skcache_entry_t)); /* clear all field */
    newone->magic[0] = MAGIC_0;
    newone->magic[1] = MAGIC_1;
    newone->oid = 0;
    newone->h1 = newone->h2 = newone->h3 = newone->h4 = -1;
    newone->ref = 0;
    newone->count = 0;
    newone->totallength = (off_t)0;
    newone->issue_time = -1;
    newone->swapout_candidate = 0;
    newone->d_recorded = 0;
    newone->m_recorded = NULL;

    return newone;
}

int
skcache_push_idle_entry(bucket_t *bk, skcache_entry_t *xen)
{
    int last;
#if 0
#ifndef NDEBUG
    Trace("idle_entry: bk# %d, %p, use %d, free %d, cup.num %d, target %p\n",
        bk->bid, bk, bk->entry_use, bk->entry_free, bk->cup.num, xen);
#endif
#endif

    last = xen->oid;

    memset(xen, 0, sizeof(skcache_entry_t));
    xen->last_oid = last;
    xen->h1 = xen->h2 = xen->h3 = xen->h4 = -1;
    xen->ref = -1;
    xen->issue_time = -1;
    xen->swapout_candidate = -1;

    xen->next = bk->entry_freelist;
    bk->entry_freelist = xen;
    bk->entry_use--;
    bk->entry_free++;

    return 0;
}


/* success 0, fail 1 */
static
int
multi_frag_lock()
{
    int c;
    int chk;

#if 1
    /* for debug */
    if(frag_use<=-50) {
        fflush(stdout);
        Error("multi_frag_lock: frag under flow, use %d, free %d",
            frag_use, frag_free);
        fflush(stderr);
        fflush(stdout);
        exit(4);
    }
#if 1
    if(z_frag_freelist_dummy[0]!=NULL) {
        fflush(stdout);
        Error("multi_frag_lock: free list barrier is broken %p",
            z_frag_freelist_dummy[0]);
        fflush(stderr);
        fflush(stdout);
        exit(4);
    }
    if(frag_dm0[0]!=SENTINEL0 || frag_dm0[1]!=SENTINEL1
        || frag_dm1[0]!=SENTINEL0 || frag_dm1[1]!=SENTINEL1 ) {
        fflush(stdout);
        Error("multi_frag_lock: SENTINEL is change <%d %d> <%d %d> expected <%d %d> <%d %d>",
            frag_dm0[0], frag_dm0[1], frag_dm1[0], frag_dm1[1],
            SENTINEL0, SENTINEL1, SENTINEL0, SENTINEL1);

        fflush(stderr);
        fflush(stdout);
        exit(4);
    }
#endif

#if 0
    if(frag_use<=10) {
        Trace("multi_frag_lock: few frags, use %d, free %d\n",
            frag_use, frag_free);
    }
#endif

#endif

    c = FRAG_NTRY;
try_lock:
    c--;
    chk = MUTEX_TRYLOCK(&frag_lock);
    if(chk==EBUSY) {
        if(c>0) {
#if 0
    Trace("multi_frag_lock: free list is busy, retry %d : u %d, f %d\n",
                c, frag_use, frag_free); 
#endif
            msleep(3);
            goto try_lock;
        }
        Error("multi_frag_lock: free list is busy <%d times>, u %d, f %d",
            c, frag_use, frag_free);
        return 1;
    }
    else
    if(chk) {
        Error("multi_frag_lock: unexpected error (%d)", chk);
        return 3;
    }

    return 0;
}


skcache_frag_t*
skcache_pull_new_frag(bucket_t *bk)
{
    skcache_frag_t* cur;
    int chk;
    int c;

#if 0
    Trace("new_frag: use %d, free %d\n",
        frag_use, frag_free);
#endif

    if(multi_frag_lock()) {
        return NULL;
    }

    if(z_frag_freelist == NULL || frag_free <= 0 ) { 
        Error("new_frag: no free item, use %d %.2f%%",
            frag_use, (double)100.0*frag_use/(frag_use+frag_free));

        MUTEX_UNLOCK(&frag_lock);
        return NULL;
    }

    cur = z_frag_freelist;
    z_frag_freelist = cur->next;
    frag_use++;
    frag_free--;
    cur->next = NULL;
    cur->len  = 0;

if(frag_free<=1) {
    Error("new_frag: left last one frag %p", z_frag_freelist);
}

    MUTEX_UNLOCK(&frag_lock);

    return cur;
}

int
RAW_skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    xf->next = z_frag_freelist;
    z_frag_freelist = xf;
    frag_use--;
    frag_free++;

    return 0;
}

int
skcache_push_idle_frag(bucket_t *bk, skcache_frag_t *xf)
{
    int chk;
    int c;

#if 0
    Trace("idle_frag: target %p, use %d, free %d\n",
        xf, frag_use, frag_free);
#endif
    if(xf==NULL) {
        return -13;
    }
    if(multi_frag_lock()) {
        return -1;
    }

    chk = RAW_skcache_push_idle_frag(bk, xf);

    MUTEX_UNLOCK(&frag_lock);

    return 0;
}


#if 0
int
check_checkpoints()
{
    int i;
    time_t now;
    bucket_t *bk;
    int width = skcache_mem_keep_min * 3;

    now = time(NULL);

    for(i=0;i<radix_h1;i++) {
        bucket_t *bk;

        bk = &skcache_bucket_base[i];

        LOCK_BK(bk);

        if(bk->entry_use >= bk->entry_lwater
            && now - bk->gc_last > width )
        {
            int chk;

    MUTEX_LOCK(&stat_lock);
    compaction_count++;
    MUTEX_UNLOCK(&stat_lock);
            chk = mark_target(bk, XSK_ENTRY);

            bk->gc_last = now + (rand() % width) + i ;  /* XXX */
        }

        UNLOCK_BK(bk);
    }

}
#endif


/*
 * re-calcurate keeptime threshould
 * argument:
 *      delta_t [msec]
 */
int
skcache_update_keeptime(int delta_t)
{
    time_t now;
    int last, keep;
    double ratio;
    int rapid_flag;
    int i;
    bucket_t *bk;
    int chk;
    char *msg;

    now = time(NULL);
    rapid_flag = 0;

    MUTEX_LOCK(&config_lock);
    ratio = sim_hist_average_d(&fratio_hist);
    last = keep = skcache_mem_keep_cur;

    if(skcache_keeptime_token>0) {
        skcache_keeptime_token--;
        goto skip_update;
    }

    /* slow up and rapid down */
    if(ratio>skcache_frag_hwater_ratio) {
        msg = "rapid-down";
        skcache_mem_keep_cur = 9*keep/10;
        skcache_keeptime_token = skcache_keeptime_token_init;
    }
    else
    if(ratio>skcache_frag_lwater_ratio) {
        msg = "no-change";
    }
    else {
        msg = "up";
        skcache_mem_keep_cur = keep + delta_t/1000 + 1;
    }

    if(skcache_mem_keep_cur<skcache_mem_keep_min) {
        Trace("  underflow min %d\n", skcache_mem_keep_min);
        skcache_mem_keep_cur = skcache_mem_keep_min;
    }
    if(skcache_mem_keep_cur>skcache_mem_keep_max) {
        Trace("  overflow  max %d\n", skcache_mem_keep_max);
        skcache_mem_keep_cur = skcache_mem_keep_max;
    }

    Trace("KT updating %d frag %.2f%% (%d) %s thre. %d -> %d\n",
        now, ratio, skcache_keeptime_token, msg, last, skcache_mem_keep_cur);

skip_update:
    MUTEX_UNLOCK(&config_lock);

    if(rapid_flag) {
        for(i=0;i<radix_h1;i++) {
            bk = &skcache_bucket_base[i];

            chk = MUTEX_TRYLOCK(&bk->mutex);
            if(chk==EBUSY) {
            }
            else {
                bk->gc_push++;
                MUTEX_UNLOCK(&bk->mutex);
            }

        }
    }
}

/*
 * When it find @ (at-mark) in prefix of cache dirctory,
 * the program recognize using many disk and its pattern.
 */
check_disk_spliting()
{
    char *p;
    int v, i;
    char tmp[L_PATH];

    v = -1;
    p = cache_dir_prefix;
    while(*p) {
        if(*p=='@') {
            v = atoi(p+1);
            *p = '\0';
            break;
        }
        p++;
    }

    if(v>0) {
        radix_disk = v;

#if 0
        Trace("cache directories are splited to %d disks\n", radix_disk);
#endif
        Trace(" Spricing: cache directories stored into %d disks\n",
            radix_disk);
        Trace("           %d directories per disk\n",
            skcache_bucket_num/radix_disk);

#if 0
        for(i=0;i<radix_disk;i++) {
            Trace("   disk# %3d: %s%d\n", i, cache_dir_prefix, i);
        }
#endif
    }
}

/*
 * initialized cache parameters and areas
 */
long long
skcache_init(time_t xmin_date, int xbk_num, int xsk_level, int xsk_len,
        int xffl_num, int xffl_size)
{
    int i,j;
    int chk;
    long long got;
    long long sum, tmp_sum;
    long long all;
    bucket_t *bk;
    time_t now;

    long long tmp;

#if 0
    Trace("skcache_init: min_date %d\n\
    bucket: number %d, level %d, length %d\n\
    frag: length %d, size %d\n",
        xmin_date, xbk_num, xsk_level, xsk_len, xffl_num, xffl_size);
#endif

    if(xmin_date>0) {
        min_date = xmin_date;
    }
    else {
#if 0
        min_date = time(NULL);
        Trace("set min_date as %d (current date/time)\n", min_date);
#endif
        Trace("set min_date as negative it is mean, use stored objects before boot\n");
    }

    now = time(NULL);

    sum = 0L;



    /*
     * NIL
     */
    NIL = (skcache_entry_t*)MALLOC(sizeof(bucket_t));
    NIL->h1 = INT_MAX;
    NIL->h2 = INT_MAX;
    NIL->h3 = INT_MAX;

    sum += sizeof(bucket_t);
    mlist_add_elem("NIL", (void*)NIL, sizeof(bucket_t));

    skcache_bucket_num = xbk_num;

#if 0
    Trace("require_radix_h2 %d, require_radix_h3 %d\n",
            require_radix_h2, require_radix_h3);
#endif
    
    /*
     * cache entries
     */
    radix_h1 = skcache_bucket_num;
    radix_h2 = require_radix_h2;
    radix_h3 = require_radix_h3;
#if 0
    radix_h4 = 513;
#endif
    radix_h4 = 7537;                /* XXX */


    if(radix_h2>RADIX_H2_MAX) {
        Error("radix_h2 (%d) is too large (max %d)\n",
            radix_h2, RADIX_H2_MAX);
    }
    Trace("\n\
Hashes:\n\
  Radixes               %d, %d, %d, %d\n",
        radix_h1, radix_h2, radix_h3, radix_h4);



    dir_num = radix_h1*radix_h2;

    if(dir_num<0) {
        Trace("dir_num %d is stange\n", dir_num);
        exit(4);
    }

#if 0
    check_disk_spliting();
#endif

#ifdef OBJECT_PACKING
    {
#if 0
        int fdpoolsize = radix_h1*radix_h2*radix_h3;
#endif
        int fdpoolsize = radix_h1*radix_h2;
        long long tsize;
        long long tminsize, tmaxsize;

        pfd_pool = (int*)MALLOC(sizeof(int)*fdpoolsize);
        ifd_pool = (int*)MALLOC(sizeof(int)*fdpoolsize);
        pos_pool = (int*)MALLOC(sizeof(int)*fdpoolsize);

    mlist_add_elem("pfd_pool",  (void*)pfd_pool, sizeof(int)*fdpoolsize);
    mlist_add_elem("ifd_pool",  (void*)ifd_pool, sizeof(int)*fdpoolsize);
    mlist_add_elem("pos_pool",  (void*)pos_pool, sizeof(int)*fdpoolsize);

        for(i=0;i<fdpoolsize;i++) {
            ifd_pool[i] = -1;
            pfd_pool[i] = -1;
            pos_pool[i] = -1;
        }

        tsize = (long long)radix_h4*OFILE_UNIT * fdpoolsize;
        tminsize = (long long)1*OFILE_UNIT * fdpoolsize;
        tmaxsize = (long long)OFILE_SLOT*OFILE_UNIT * fdpoolsize;

        Trace("\
Object Packing:\n\
  OFILE_UNIT        %7d\n\
  OFILE_SLOT        %7d\n\
  OFILE_TRY         %7d\n\
\n\
  appear opportunity per cell %d\n\
\n\
  the number of files = %d\n\
  file   %12lld (%7.2fM) .. %12lld (%7.2fM)\n\
  total  %12lld (%7.2fG) .. %12lld (%7.2fG)\n\
\n\n",
            OFILE_UNIT, OFILE_SLOT, OFILE_TRY,
            radix_h4/OFILE_SLOT,
            fdpoolsize,
            (long long)1*OFILE_UNIT, fdM((long)1*OFILE_UNIT),
            (long long)OFILE_SLOT*OFILE_UNIT, fdM((long)OFILE_SLOT*OFILE_UNIT),
            (long long)tminsize, fdG(tminsize),
            (long long)tmaxsize, fdG(tmaxsize));

    }
#endif


#if 0

    dir_size_max    = require_dir_size;
    dir_size_hwater = require_dir_size*9/10;
    dir_size_lwater = require_dir_size*6/10;

#else

    dir_size_max    = OFILE_UNIT*OFILE_SLOT * 11/10;
    dir_size_hwater = dir_size_max*19/20;
    dir_size_lwater = dir_size_max*16/20;

#endif


#if 1
    tmp = ((long long) dir_size_max * dir_num);

    Trace("\n\
Disk:\n\
  The number of directories      %d\n\
  Max size of each directory     %12lld %9.2fK %5.2fM \n\
    High water                       %6.2f%% : %12lld %9.2fK %5.2fM \n\
    Low water                        %6.2f%% : %12lld %9.2fK %5.2fM \n\
  Actual cache size on disk      %12lld %9.2fM %5.2fG\n",
        dir_num,
        (long long) dir_size_max, fdK(dir_size_max), fdM(dir_size_max),
        ((double) dir_size_hwater*100.0)/dir_size_max,
        (long long) dir_size_hwater, fdK(dir_size_hwater), fdM(dir_size_hwater),
        ((double) dir_size_lwater*100.0)/dir_size_max,
        (long long) dir_size_lwater, fdK(dir_size_lwater), fdM(dir_size_lwater),
        tmp, fdM(tmp), fdG(tmp));


    if(radix_disk>2) {
        Trace("\
    Splitted to %3d disks            %.2fG per disk\n",
        radix_disk, fdG(tmp/radix_disk) );
    }
#endif



    dir_lock = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_lock",   (void*)dir_lock, sizeof(int)*dir_num);

#if 1
    Trace("\n\
Signatures for directories:\n\
  SIZE_OF_UINT        %d\n\
  DSIG_BIT_WIDTH      %d\n\
  DSIG_VALUE_RANGE    %d\n\
  DSIG_VALUE_RES      %d\n\
  DSIG_ELEM_WIDTH     %d\n\
  DSIG_BYTE_WIDTH     %d\n",
        SIZE_OF_UINT,
        DSIG_BIT_WIDTH, DSIG_VALUE_RANGE, DSIG_VALUE_RES,
        DSIG_ELEM_WIDTH, DSIG_BYTE_WIDTH);
    Trace("\n");
#endif

    dir_sig = (u_int32_t*)MALLOC(sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);
    sum += sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH;
    mlist_add_elem("dir_sig",   (void*)dir_sig, sizeof(u_int32_t)*dir_num*DSIG_ELEM_WIDTH);

    dir_sz  = (off_t*)MALLOC(sizeof(off_t)*dir_num);
    sum += sizeof(off_t)*dir_num;
    mlist_add_elem("dir_sz",    (void*)dir_sz, sizeof(off_t)*dir_num);

    dir_count = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_count", (void*)dir_count, sizeof(int)*dir_num);

    dir_time = (int*)MALLOC(sizeof(int)*dir_num);
    sum += sizeof(int)*dir_num;
    mlist_add_elem("dir_time",  (void*)dir_time, sizeof(int)*dir_num);

    for(i=0;i<dir_num;i++) {
#if 0
        dir_sig[i]      = (int64_t)-1L;
#endif
        DSIG_ALLSET(i);
        dir_sz[i]       = (off_t)-1L;
        dir_count[i]    = -1;
        dir_time[i]     = now - rand()%skcache_signature_refresh;
        dir_lock[i]     = 0;
    }





    skcache_bucket_base = (bucket_t*)MALLOC(sizeof(bucket_t)*xbk_num);
    if(skcache_bucket_base==NULL) {
        Error("not enough memory for skcache_bucket_base");
        skcache_bucket_num = 0;
        return -1;
    }

    sum += sizeof(bucket_t)*xbk_num;
    mlist_add_elem("bucket bases",
        (void*)skcache_bucket_base, sizeof(bucket_t)*xbk_num);

    tmp_sum = 0LL;

    for(i=0;i<skcache_bucket_num;i++) {
#if 0
Trace("bk# %d %p\n", i, skcache_bucket_base[i]);
#endif

        chk = bk_new(&skcache_bucket_base[i], xsk_level, xsk_len, NIL);
        if(chk==0) {
            Error("initialize error in bucket# %d", i);
            return -2;
        }
        tmp_sum += chk;

        bk = &skcache_bucket_base[i];
        bk->bid = i;
        bk->own = '.';

        got = alloc_entry_freelist(&bk->entry_freelist, xsk_len+1);
        if(got<=0) {
            Error("initialize error in object entry elem_freelist# %d", i);
            return -2;
        }
        else {
            tmp_sum += got;
        }

        bk->entry_all  = xsk_len+1;
        bk->entry_use  = 0;
        bk->entry_free = xsk_len+1;
#if 0
        bk->entry_hhwater   = (int) (98.0 * bk->entry_all/100);
        bk->entry_hwater    = (int) (92.0 * bk->entry_all/100);
        bk->entry_lwater    = (int) (69.0 * bk->entry_all/100);
#endif
#if 1
        bk->entry_hhwater   = 98*bk->entry_all/100 - rand()%(bk->entry_all/20); 
        bk->entry_hwater    = 82*bk->entry_all/100 - rand()%(bk->entry_all/20); 
        bk->entry_lwater    = 69*bk->entry_all/100 - rand()%(bk->entry_all/20); 
#endif

#if 0
    Trace("bk# %4d: %5d %5d->%5d %5d->%5d\n",
        bk->bid, bk->entry_all,
        80*bk->entry_all/100, bk->entry_hwater,
        60*bk->entry_all/100, bk->entry_lwater
        );
#endif

#if 0
#ifndef NDEBUG
        bk->entry_first = (skcache_entry_t*) ((char*)bk->entry_freelist
                    + sizeof(skcache_entry_t)* 0);
        bk->entry_last = (skcache_entry_t*) ((char*)bk->entry_freelist
                    + sizeof(skcache_entry_t)* (xsk_len+1));
#endif

#ifndef NDEBUG
#if 1
    Trace(" %3d head of freelist %p first %p last %p\n",
        i, bk->entry_freelist, bk->entry_first, bk->entry_last);
    Trace("     2nd of freelist %p next of first %p\n",
        bk->entry_freelist->next, 
        (char*)bk->entry_freelist + sizeof(skcache_entry_t)* 1
    );
#endif
#endif
#endif

        bk->frag_use = 0;

        bk->gc_last = -1;
        bk->gc_push = 0; 
        bk->gc_rest = 0;
        bk->gc_checkpoint = 
            now - skcache_mem_keep_cur/4 + (rand()% (skcache_mem_keep_cur/2));
        bk->gc_ntarget = 0;

        bk->dirpos = rand()%radix_h2;
        bk->dirty2 = -1;
        bk->dirty3 = -1;

    }

#if 1
    Trace("Memory:\n\
 entries:\n\
  the number of entries %d (/bucket)\n\
  watermark             hhigh %d (%.2f%%), high %d (%.2f%%), low %d (%.2f%%)\n",
            bk->entry_all,
            bk->entry_hhwater,  ((double)bk->entry_hhwater*100)/bk->entry_all,
            bk->entry_hwater,   ((double)bk->entry_hwater*100)/bk->entry_all,
            bk->entry_lwater,   ((double)bk->entry_lwater*100)/bk->entry_all
        );
#endif

    sum += tmp_sum;
    mlist_add_elem("bucket bodies", (void*)NULL, tmp_sum);


    /*
     * frags
     */
    got = alloc_frag_freelist(&z_frag_freelist, 
            (xffl_num+1)*skcache_bucket_num, xffl_size);
    if(got<0) {
        Error("Sorry, no memory for frag free list");
        return -2;
    }

    frag_free = (xffl_num+1)*skcache_bucket_num;
    frag_use  = 0;

#if 1
    frag_hwater = 16*frag_free/20;
    frag_lwater = 12*frag_free/20;
#endif
    skcache_frag_hwater_ratio = 80.0;
    skcache_frag_lwater_ratio = 60.0;

    sum += got;
    mlist_add_elem("frags", (void*)z_frag_freelist, got);

#if 1
    Trace(" frags:\n\
  the number of frags   %d\n\
  watermark             high %.2f%%, low %.2f%%\n",
        frag_free, skcache_frag_hwater_ratio, skcache_frag_lwater_ratio);
#if 1
    Trace("\
  Size of frag cell is %d+%d = %d\n",
        sizeof(skcache_frag_t), xffl_size, sizeof(skcache_frag_t)+xffl_size);
#endif
#endif







    skcache_keeptime_token_init = require_keeptime_token_val;
    if(skcache_keeptime_token_init<2) {
        skcache_keeptime_token_init = 2;
    }
    skcache_keeptime_token = skcache_keeptime_token_init;

    Trace("  keeptime updating token    %d\n", skcache_keeptime_token_init);
    Trace("  keeptime updating interval %d\n", skcache_keeptime_token_init*2);

    Trace("\n\
Magic:\n\
  decimal   %3d %3d\n\
  hex       %3x %3x\n\
  charactor '%c%c'\n", 
    MAGIC_0, MAGIC_1, MAGIC_0, MAGIC_1, MAGIC_0, MAGIC_1);



    swapout_token = swapout_token_init = require_swap_token_val;
    swapout_token_interval = require_swap_token_interval;


    Trace("swapout token %d in %d msec -> rate %.4f swapout/sec\n",
        require_swap_token_val, require_swap_token_interval,
        (double)require_swap_token_val*
            (1000.0/(double)require_swap_token_interval)
        );

    disk_sweep_interval = require_disk_sweep_interval;

#if 0
    Trace("skcache_init: allocate %lld byte(s) %dK %.2fM %.2fG\n",
        (long long) sum, (long)dK(sum), fdM(sum), fdG(sum));
#endif

    sim_stat_alloc_new(&swapload_time,  "swapload time [msec]",
        K(1),   swapload_time_count);
    sim_stat_alloc_new(&swapin_time,    "swapin time [msec]",
        K(1),   swapin_time_count);
    sim_stat_alloc_new(&swapout_time,   "swapout time [msec]",
        K(1),   swapout_time_count);

    sim_stat_U_alloc_new(&compaction_time,  "compaction time [msec]",
        100, K(2),   compaction_time_count);
    sim_stat_U_alloc_new(&checkdir_time,    "checkdir time [msec]",
        10, K(2),   checkdir_time_count);
    sim_stat_U_alloc_new(&checkdir_time_short,    "checkdir time (short) [msec]",
        10, K(2),   checkdir_time_short_stat_count);
    sim_stat_U_alloc_new(&checkmem_time,    "checkmem time [msec]",
        100, K(2),   checkmem_time_count);

    sim_stat_alloc_new(&swapout_fail_reason,    "swapout fail reason",
        K(1),   swapout_fail_reason_count);
    sim_stat_alloc_new(&swapload_fail_reason,   "swapload fail reason",
        K(1),   swapload_fail_reason_count);

    sim_hist_alloc_new(&fratio_hist,            "frag ratio history",
        skcache_keeptime_token_init*2,  fratio_hist_area);

    return sum;
}

